# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import List, Optional, Tuple

import torch
import torch.nn as nn

from src.gluonts.core.component import validated
from pts.modules import RealNVP, MAF, FlowOutput, MeanScaler, NOPScaler


class TransformerTempFlowTrainingNetwork(nn.Module):
    @validated()
    def __init__(
        self,
        input_size: int,
        d_model: int,
        num_heads: int,
        act_type: str,
        dropout_rate: float,
        dim_feedforward_scale: int,
        num_encoder_layers: int,
        num_decoder_layers: int,
        history_length: int,
        context_length: int,
        prediction_length: int,
        lags_seq: List[int],
        target_dim: int,
        conditioning_length: int,
        flow_type: str,
        n_blocks: int,
        hidden_size: int,
        n_hidden: int,
        dequantize: bool,
        scaling: bool = True,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.target_dim = target_dim
        self.prediction_length = prediction_length
        self.context_length = context_length
        self.history_length = history_length
        self.scaling = scaling

        assert len(set(lags_seq)) == len(
            lags_seq
        ), "no duplicated lags allowed!"
        lags_seq.sort()
        self.lags_seq = lags_seq

        self.encoder_input = nn.Linear(input_size, d_model)
        self.decoder_input = nn.Linear(input_size, d_model)

        # [B, T, d_model] where d_model / num_heads is int
        self.transformer = nn.Transformer(
            d_model=d_model,
            nhead=num_heads,
            num_encoder_layers=num_encoder_layers,
            num_decoder_layers=num_decoder_layers,
            dim_feedforward=dim_feedforward_scale * d_model,
            dropout=dropout_rate,
            activation=act_type,
        )

        flow_cls = {
            "RealNVP": RealNVP,
            "MAF": MAF,
        }[flow_type]
        self.flow = flow_cls(
            input_size=target_dim,
            n_blocks=n_blocks,
            n_hidden=n_hidden,
            hidden_size=hidden_size,
            cond_label_size=conditioning_length,
        )
        self.dequantize = dequantize

        self.distr_output = FlowOutput(
            self.flow, input_size=target_dim, cond_size=conditioning_length
        )

        self.proj_dist_args = self.distr_output.get_args_proj(d_model)

        self.embed_dim = 1
        self.embed = nn.Embedding(
            num_embeddings=self.target_dim, embedding_dim=self.embed_dim
        )

        if self.scaling:
            self.scaler = MeanScaler(keepdim=True)
        else:
            self.scaler = NOPScaler(keepdim=True)

        # mask
        self.register_buffer(
            "tgt_mask",
            self.transformer.generate_square_subsequent_mask(
                prediction_length
            ),
        )

    @staticmethod
    def get_lagged_subsequences(
        sequence: torch.Tensor,
        sequence_length: int,
        indices: List[int],
        subsequences_length: int = 1,
    ) -> torch.Tensor:
        """
        Returns lagged subsequences of a given sequence.
        Parameters
        ----------
        sequence
            the sequence from which lagged subsequences should be extracted.
            Shape: (N, T, C).
        sequence_length
            length of sequence in the T (time) dimension (axis = 1).
        indices
            list of lag indices to be used.
        subsequences_length
            length of the subsequences to be extracted.
        Returns
        --------
        lagged : Tensor
            a tensor of shape (N, S, C, I),
            where S = subsequences_length and I = len(indices),
            containing lagged subsequences.
            Specifically, lagged[i, :, j, k] = sequence[i, -indices[k]-S+j, :].
        """
        # we must have: history_length + begin_index >= 0
        # that is: history_length - lag_index - sequence_length >= 0
        # hence the following assert
        assert max(indices) + subsequences_length <= sequence_length, (
            f"lags cannot go further than history length, found lag "
            f"{max(indices)} while history length is only {sequence_length}"
        )
        assert all(lag_index >= 0 for lag_index in indices)

        lagged_values = []
        for lag_index in indices:
            begin_index = -lag_index - subsequences_length
            end_index = -lag_index if lag_index > 0 else None
            lagged_values.append(
                sequence[:, begin_index:end_index, ...].unsqueeze(1)
            )
        return torch.cat(lagged_values, dim=1).permute(0, 2, 3, 1)

    def create_network_input(
        self,
        past_time_feat: torch.Tensor,
        past_target_cdf: torch.Tensor,
        past_observed_values: torch.Tensor,
        past_is_pad: torch.Tensor,
        future_time_feat: Optional[torch.Tensor],
        future_target_cdf: Optional[torch.Tensor],
        target_dimension_indicator: torch.Tensor,
    ) -> Tuple[
        torch.Tensor,
        torch.Tensor,
        torch.Tensor,
    ]:
        """
        Unrolls the RNN encoder over past and, if present, future data. Returns
        outputs and state of the encoder, plus the scale of past_target_cdf and
        a vector of static features that was constructed and fed as input to
        the encoder. All tensor arguments should have NTC layout.

        Parameters
        ----------
        past_time_feat
            Past time features (batch_size, history_length, num_features)
        past_target_cdf
            Past marginal CDF transformed target values (batch_size,
            history_length, target_dim)
        past_observed_values
            Indicator whether or not the values were observed (batch_size,
            history_length, target_dim)
        past_is_pad
            Indicator whether the past target values have been padded
            (batch_size, history_length)
        future_time_feat
            Future time features (batch_size, prediction_length, num_features)
        future_target_cdf
            Future marginal CDF transformed target values (batch_size,
            prediction_length, target_dim)
        target_dimension_indicator
            Dimensionality of the time series (batch_size, target_dim)

        Returns
        -------
        outputs
            RNN outputs (batch_size, seq_len, num_cells)
        states
            RNN states. Nested list with (batch_size, num_cells) tensors with
        dimensions target_dim x num_layers x (batch_size, num_cells)
        scale
            Mean scales for the time series (batch_size, 1, target_dim)
        lags_scaled
            Scaled lags(batch_size, sub_seq_len, target_dim, num_lags)
        inputs
            inputs to the RNN
        """

        past_observed_values = torch.min(
            past_observed_values, 1 - past_is_pad.unsqueeze(-1)
        )

        if future_time_feat is None or future_target_cdf is None:
            time_feat = past_time_feat[:, -self.context_length :, ...]
            sequence = past_target_cdf
            sequence_length = self.history_length
            subsequences_length = self.context_length
        else:
            time_feat = torch.cat(
                (
                    past_time_feat[:, -self.context_length :, ...],
                    future_time_feat,
                ),
                dim=1,
            )
            sequence = torch.cat((past_target_cdf, future_target_cdf), dim=1)
            sequence_length = self.history_length + self.prediction_length
            subsequences_length = self.context_length + self.prediction_length

        # (batch_size, sub_seq_len, target_dim, num_lags)
        lags = self.get_lagged_subsequences(
            sequence=sequence,
            sequence_length=sequence_length,
            indices=self.lags_seq,
            subsequences_length=subsequences_length,
        )

        # scale is computed on the context length last units of the past target
        # scale shape is (batch_size, 1, target_dim)
        _, scale = self.scaler(
            past_target_cdf[:, -self.context_length :, ...],
            past_observed_values[:, -self.context_length :, ...],
        )

        # (batch_size, sub_seq_len, target_dim, num_lags)
        lags_scaled = lags / scale.unsqueeze(-1)

        # assert_shape(
        #     lags_scaled, (-1, unroll_length, self.target_dim, len(self.lags_seq)),
        # )

        input_lags = lags_scaled.reshape(
            (-1, subsequences_length, len(self.lags_seq) * self.target_dim)
        )

        # (batch_size, target_dim, embed_dim)
        index_embeddings = self.embed(target_dimension_indicator)
        # assert_shape(index_embeddings, (-1, self.target_dim, self.embed_dim))

        # (batch_size, seq_len, target_dim * embed_dim)
        repeated_index_embeddings = (
            index_embeddings.unsqueeze(1)
            .expand(-1, subsequences_length, -1, -1)
            .reshape(
                (-1, subsequences_length, self.target_dim * self.embed_dim)
            )
        )

        # (batch_size, sub_seq_len, input_dim)
        inputs = torch.cat(
            (input_lags, repeated_index_embeddings, time_feat), dim=-1
        )

        return inputs, scale, index_embeddings

    def distr_args(self, decoder_output: torch.Tensor):
        """
        Returns the distribution of DeepVAR with respect to the RNN outputs.

        Parameters
        ----------
        rnn_outputs
            Outputs of the unrolled RNN (batch_size, seq_len, num_cells)
        scale
            Mean scale for each time series (batch_size, 1, target_dim)

        Returns
        -------
        distr
            Distribution instance
        distr_args
            Distribution arguments
        """
        (distr_args,) = self.proj_dist_args(decoder_output)

        # # compute likelihood of target given the predicted parameters
        # distr = self.distr_output.distribution(distr_args, scale=scale)

        # return distr, distr_args
        return distr_args

    def forward(
        self,
        target_dimension_indicator: torch.Tensor,
        past_time_feat: torch.Tensor,
        past_target_cdf: torch.Tensor,
        past_observed_values: torch.Tensor,
        past_is_pad: torch.Tensor,
        future_time_feat: torch.Tensor,
        future_target_cdf: torch.Tensor,
        future_observed_values: torch.Tensor,
    ) -> Tuple[torch.Tensor, ...]:
        """
        Computes the loss for training DeepVAR, all inputs tensors representing
        time series have NTC layout.

        Parameters
        ----------
        target_dimension_indicator
            Indices of the target dimension (batch_size, target_dim)
        past_time_feat
            Dynamic features of past time series (batch_size, history_length,
            num_features)
        past_target_cdf
            Past marginal CDF transformed target values (batch_size,
            history_length, target_dim)
        past_observed_values
            Indicator whether or not the values were observed (batch_size,
            history_length, target_dim)
        past_is_pad
            Indicator whether the past target values have been padded
            (batch_size, history_length)
        future_time_feat
            Future time features (batch_size, prediction_length, num_features)
        future_target_cdf
            Future marginal CDF transformed target values (batch_size,
            prediction_length, target_dim)
        future_observed_values
            Indicator whether or not the future values were observed
            (batch_size, prediction_length, target_dim)

        Returns
        -------
        distr
            Loss with shape (batch_size, 1)
        likelihoods
            Likelihoods for each time step
            (batch_size, context + prediction_length, 1)
        distr_args
            Distribution arguments (context + prediction_length,
            number_of_arguments)
        """

        # seq_len = self.context_length + self.prediction_length

        # unroll the decoder in "training mode", i.e. by providing future data
        # as well
        inputs, scale, _ = self.create_network_input(
            past_time_feat=past_time_feat,
            past_target_cdf=past_target_cdf,
            past_observed_values=past_observed_values,
            past_is_pad=past_is_pad,
            future_time_feat=future_time_feat,
            future_target_cdf=future_target_cdf,
            target_dimension_indicator=target_dimension_indicator,
        )

        enc_inputs = inputs[:, : self.context_length, ...]
        dec_inputs = inputs[:, self.context_length :, ...]

        enc_out = self.transformer.encoder(
            self.encoder_input(enc_inputs).permute(1, 0, 2)
        )

        dec_output = self.transformer.decoder(
            self.decoder_input(dec_inputs).permute(1, 0, 2),
            enc_out,
            tgt_mask=self.tgt_mask,
        )

        if self.scaling:
            self.flow.scale = scale

        # we sum the last axis to have the same shape for all likelihoods
        # (batch_size, subseq_length, 1)
        if self.dequantize:
            future_target_cdf += torch.rand_like(future_target_cdf)

        distr_args = self.distr_args(
            decoder_output=dec_output.permute(1, 0, 2)
        )
        # likelihoods = -self.flow.log_prob(target, distr_args).unsqueeze(-1)
        loss = -self.flow.log_prob(future_target_cdf, distr_args).unsqueeze(-1)

        # # assert_shape(likelihoods, (-1, seq_len, 1))

        # past_observed_values = torch.min(
        #     past_observed_values, 1 - past_is_pad.unsqueeze(-1)
        # )

        # # (batch_size, subseq_length, target_dim)
        # observed_values = torch.cat(
        #     (
        #         past_observed_values[:, -self.context_length :, ...],
        #         future_observed_values,
        #     ),
        #     dim=1,
        # )

        # # mask the loss at one time step if one or more observations is missing
        # # in the target dimensions (batch_size, subseq_length, 1)
        # loss_weights, _ = observed_values.min(dim=-1, keepdim=True)

        # # assert_shape(loss_weights, (-1, seq_len, 1))

        # loss = weighted_average(likelihoods, weights=loss_weights, dim=1)

        # assert_shape(loss, (-1, -1, 1))

        # self.distribution = distr

        # return (loss.mean(), likelihoods, distr_args)
        return loss.mean(), distr_args


class TransformerTempFlowPredictionNetwork(TransformerTempFlowTrainingNetwork):
    def __init__(self, num_parallel_samples: int, **kwargs) -> None:
        super().__init__(**kwargs)
        self.num_parallel_samples = num_parallel_samples

        # for decoding the lags are shifted by one,
        # at the first time-step of the decoder a lag of one corresponds to
        # the last target value
        self.shifted_lags = [l - 1 for l in self.lags_seq]

    def sampling_decoder(
        self,
        past_target_cdf: torch.Tensor,
        target_dimension_indicator: torch.Tensor,
        time_feat: torch.Tensor,
        scale: torch.Tensor,
        enc_out: torch.Tensor,
    ) -> torch.Tensor:
        """
        Computes sample paths by unrolling the RNN starting with a initial
        input and state.

        Parameters
        ----------
        past_target_cdf
            Past marginal CDF transformed target values (batch_size,
            history_length, target_dim)
        target_dimension_indicator
            Indices of the target dimension (batch_size, target_dim)
        time_feat
            Dynamic features of future time series (batch_size, history_length,
            num_features)
        scale
            Mean scale for each time series (batch_size, 1, target_dim)
        begin_states
            List of initial states for the RNN layers (batch_size, num_cells)
        Returns
        --------
        sample_paths : Tensor
            A tensor containing sampled paths. Shape: (1, num_sample_paths,
            prediction_length, target_dim).
        """

        def repeat(tensor, dim=0):
            return tensor.repeat_interleave(
                repeats=self.num_parallel_samples, dim=dim
            )

        # blows-up the dimension of each tensor to
        # batch_size * self.num_sample_paths for increasing parallelism
        repeated_past_target_cdf = repeat(past_target_cdf)
        repeated_time_feat = repeat(time_feat)
        repeated_scale = repeat(scale)
        if self.scaling:
            self.flow.scale = repeated_scale
        repeated_target_dimension_indicator = repeat(
            target_dimension_indicator
        )
        repeated_enc_out = repeat(enc_out, dim=1)

        future_samples = []

        # for each future time-units we draw new samples for this time-unit
        # and update the state
        for k in range(self.prediction_length):
            lags = self.get_lagged_subsequences(
                sequence=repeated_past_target_cdf,
                sequence_length=self.history_length + k,
                indices=self.shifted_lags,
                subsequences_length=1,
            )

            lags_scaled = lags / repeated_scale.unsqueeze(-1)

            input_lags = lags_scaled.reshape(
                (-1, 1, len(self.lags_seq) * self.target_dim)
            )

            # (batch_size, target_dim, embed_dim)
            index_embeddings = self.embed(repeated_target_dimension_indicator)
            # assert_shape(index_embeddings, (-1, self.target_dim, self.embed_dim))

            # (batch_size, seq_len, target_dim * embed_dim)
            repeated_index_embeddings = (
                index_embeddings.unsqueeze(1)
                .expand(-1, 1, -1, -1)
                .reshape((-1, 1, self.target_dim * self.embed_dim))
            )

            # (batch_size, sub_seq_len, input_dim)
            dec_input = torch.cat(
                (
                    input_lags,
                    repeated_index_embeddings,
                    repeated_time_feat[:, k : k + 1, ...],
                ),
                dim=-1,
            )

            dec_output = self.transformer.decoder(
                self.decoder_input(dec_input).permute(1, 0, 2),
                repeated_enc_out,
            )

            distr_args = self.distr_args(
                decoder_output=dec_output.permute(1, 0, 2)
            )

            # (batch_size, 1, target_dim)
            new_samples = self.flow.sample(cond=distr_args)

            # (batch_size, seq_len, target_dim)
            future_samples.append(new_samples)
            repeated_past_target_cdf = torch.cat(
                (repeated_past_target_cdf, new_samples), dim=1
            )

        # (batch_size * num_samples, prediction_length, target_dim)
        samples = torch.cat(future_samples, dim=1)

        # (batch_size, num_samples, prediction_length, target_dim)
        return samples.reshape(
            (
                -1,
                self.num_parallel_samples,
                self.prediction_length,
                self.target_dim,
            )
        )

    def forward(
        self,
        target_dimension_indicator: torch.Tensor,
        past_time_feat: torch.Tensor,
        past_target_cdf: torch.Tensor,
        past_observed_values: torch.Tensor,
        past_is_pad: torch.Tensor,
        future_time_feat: torch.Tensor,
    ) -> torch.Tensor:
        """
        Predicts samples given the trained DeepVAR model.
        All tensors should have NTC layout.
        Parameters
        ----------
        target_dimension_indicator
            Indices of the target dimension (batch_size, target_dim)
        past_time_feat
            Dynamic features of past time series (batch_size, history_length,
            num_features)
        past_target_cdf
            Past marginal CDF transformed target values (batch_size,
            history_length, target_dim)
        past_observed_values
            Indicator whether or not the values were observed (batch_size,
            history_length, target_dim)
        past_is_pad
            Indicator whether the past target values have been padded
            (batch_size, history_length)
        future_time_feat
            Future time features (batch_size, prediction_length, num_features)

        Returns
        -------
        sample_paths : Tensor
            A tensor containing sampled paths (1, num_sample_paths,
            prediction_length, target_dim).

        """

        # mark padded data as unobserved
        # (batch_size, target_dim, seq_len)
        past_observed_values = torch.min(
            past_observed_values, 1 - past_is_pad.unsqueeze(-1)
        )

        inputs, scale, static_feat = self.create_network_input(
            past_time_feat=past_time_feat,
            past_target_cdf=past_target_cdf,
            past_observed_values=past_observed_values,
            past_is_pad=past_is_pad,
            future_time_feat=None,
            future_target_cdf=None,
            target_dimension_indicator=target_dimension_indicator,
        )

        enc_out = self.transformer.encoder(
            self.encoder_input(inputs).permute(1, 0, 2)
        )

        return self.sampling_decoder(
            past_target_cdf=past_target_cdf,
            target_dimension_indicator=target_dimension_indicator,
            time_feat=future_time_feat,
            scale=scale,
            enc_out=enc_out,
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from abc import ABC, abstractmethod
import torch
from torch import nn
from torch.nn.utils.rnn import pack_padded_sequence

from meta.data.batch import SeriesBatch
from .tcn import CausalCNNEncoder


class QueryEncoder(nn.Module, ABC):
    """
    Base class for query encoders.
    """

    @abstractmethod
    def forward(self, query: SeriesBatch) -> SeriesBatch:
        """
        Encodes each query with a vector of fixed size.

        Parameters
        ----------
        query:
            A SeriesBatch containing sequences of size [batch, sequence length, n_features].

        Returns
        -------
        SeriesBatch:
            The transformed batch containing sequences with the encoded queries
            of size [batch, encoding size].
        """


class LSTMQueryEncoder(QueryEncoder):
    """
    Encodes queries via the last hidden state of a uni-directional LSTM.

    Parameters
    ----------
    input_size: The number of features of the time series.
    hidden_size: The size of hidden states of LSTM, corresponds to encoding size of the queries,
        i.e. the encoder returns a tensor with shape [batch, hidden_size].
    num_layers: number of layers of the LSTM.
    """

    def __init__(self, input_size: int, hidden_size: int, num_layers: int):
        super().__init__()
        self.encoder = nn.LSTM(
            input_size=input_size,
            hidden_size=hidden_size,
            num_layers=num_layers,
            bidirectional=False,
            batch_first=True,
        )

    def forward(self, query: SeriesBatch) -> torch.Tensor:
        query_packed = pack_padded_sequence(
            query.sequences,
            query.lengths.cpu(),
            batch_first=True,
            enforce_sorted=False,
        )
        _, (output, _) = self.encoder(query_packed)
        # return last layer of hidden states
        return SeriesBatch(output[-1, ...], None, query.split_sections)


class TcnQueryEncoder(QueryEncoder):
    """
    Encodes queries via the last output of a WaveNet.

    Parameters
    ----------
    input_size: The number of features of the time series.
    hidden_size: The size of hidden states of LSTM, corresponds to encoding size of the queries,
        i.e. the encoder returns a tensor with shape [batch, hidden_size].
    num_layers: number of layers of the LSTM.
    """

    def __init__(
        self,
        encoder: nn.Module = None,
        num_channels: int = 64,
        num_layers: int = 5,
        kernel_size: int = 2,
    ):
        super().__init__()
        if encoder:
            self.encoder = encoder
        else:
            self.encoder = CausalCNNEncoder(
                in_channels=1,
                channels=num_channels,
                depth=num_layers,
                out_channels=num_channels,
                kernel_size=kernel_size,
            )

    def forward(self, query: SeriesBatch) -> torch.Tensor:
        output = self.encoder(query.sequences.permute(0, 2, 1))

        # return the last state
        idx = query.lengths - 1
        return SeriesBatch(
            output[torch.range(0, len(idx) - 1, dtype=torch.long), :, idx],
            None,
            query.split_sections,
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import Dict, List, Optional, Tuple

import mxnet as mx
import numpy as np

from src.gluonts.core.component import Type, validated
from src.gluonts.itertools import prod
from src.gluonts.mx.model.deepar import DeepAREstimator
from src.gluonts.mx.model.deepar._network import DeepARPredictionNetwork
from src.gluonts.mx.model.deepvar_hierarchical._estimator import projection_mat
from src.gluonts.mx.model.deepvar_hierarchical._network import coherency_error
from src.gluonts.mx.distribution import Distribution, EmpiricalDistribution
from src.gluonts.mx import Tensor
from src.gluonts.mx.distribution import TransformedPiecewiseLinear

from src.gluonts.nursery.temporal_hierarchical_forecasting.utils import utils
from src.gluonts.nursery.temporal_hierarchical_forecasting.model.cop_deepar import (
    gluonts_fixes,
    gnn,
)


def reconcile_samples(
    reconciliation_mat: Tensor,
    samples: Tensor,
    non_negative: bool = False,
    num_iters: int = 10,
) -> Tensor:
    if not non_negative:
        return mx.nd.dot(samples, reconciliation_mat, transpose_b=True)
    else:
        # Dykstra's projection method: Projection onto the intersection of convex sets.
        x = samples
        p = mx.nd.zeros_like(x)
        q = mx.nd.zeros_like(x)
        for _ in range(num_iters):
            # Projection onto the non-negative orthant.
            y = mx.nd.relu(x + p)
            p = x + p - y

            # Projection onto the null space.
            x = mx.nd.dot(y + q, reconciliation_mat, transpose_b=True)
            q = y + q - x

        return x


class COPNetwork(mx.gluon.HybridBlock):
    @validated()
    def __init__(
        self,
        estimators: List[DeepAREstimator],
        prediction_length: int,
        temporal_hierarchy: utils.TemporalHierarchy,
        do_reconciliation: bool,
        dtype: Type,
        use_gnn: bool,
        use_mlp: bool,
        adj_mat_option: str,
        non_negative: bool = False,
        naive_reconciliation: bool = False,
        prediction: bool = False,
        loss_function: str = "crps_univariate",
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.prediction_length = prediction_length
        self.temporal_hierarchy = temporal_hierarchy
        self.use_gnn = use_gnn
        self.use_mlp = use_mlp
        self.adj_mat_option = adj_mat_option
        self.do_reconciliation = do_reconciliation
        self.non_negative = non_negative
        self.loss_function = loss_function
        self.dtype = dtype

        if naive_reconciliation:
            M = utils.naive_reconcilation_mat(
                self.temporal_hierarchy.agg_mat, self.temporal_hierarchy.nodes
            )
        else:
            M = projection_mat(S=self.temporal_hierarchy.agg_mat)
        self.M = mx.nd.array(M)

        self.estimators = estimators

        self.models = []
        with self.name_scope():
            for estimator in estimators:
                if not prediction:
                    self.network = estimator.create_training_network()
                else:
                    self.network = gluonts_fixes.create_prediction_network(
                        estimator
                    )

                self.register_child(self.network)
                self.models.append(self.network)

            if self.use_gnn:
                # GNN Layer: Do message passing for `L-1` times, where `L` is the number of levels of the hierarchy.
                self.gnn = gnn.GNN(
                    units=self.estimators[0].num_cells,
                    num_layers=len(self.temporal_hierarchy.agg_multiples) - 1,
                    adj_matrix=mx.nd.array(
                        self.temporal_hierarchy.adj_mat(
                            option=self.adj_mat_option
                        )
                    ),
                    use_mlp=self.use_mlp,
                )

    def get_target_related_feat_at_agg_level(
        self,
        agg_level: int,
        past_target: Tensor,
        past_observed_values: Tensor,
        past_is_pad: Tensor,
        future_target: Optional[Tensor] = None,
        future_observed_values: Optional[Tensor] = None,
    ) -> Dict:
        """
        Aggregate target at the given aggregate level along with updating
        observed value and pad indicators.

        :param agg_level:
        :param past_target:
        :param past_observed_values:
        :param past_is_pad:
        :param future_target:
        :param future_observed_values:
        :return:
        """
        agg_multiple = self.temporal_hierarchy.agg_multiples[agg_level]

        # Truncating the history length of the base time series to the nearest multiple.
        base_history_length = (
            past_target.shape[1] // agg_multiple
        ) * agg_multiple

        past_target_agg = (
            utils.agg_series(
                past_target.slice_axis(
                    axis=1, begin=-base_history_length, end=None
                ),
                agg_multiple=agg_multiple,
            )
            .squeeze(axis=-1)
            .slice_axis(
                axis=1, begin=-self.models[agg_level].history_length, end=None
            )
        )

        past_is_pad_agg = (
            utils.agg_series(
                past_is_pad.slice_axis(
                    axis=1, begin=-base_history_length, end=None
                ),
                agg_multiple=agg_multiple,
            )
            .squeeze(axis=-1)
            .slice_axis(
                axis=1, begin=-self.models[agg_level].history_length, end=None
            )
        )

        past_is_pad_agg = mx.nd.where(
            past_is_pad_agg == 0.0,
            mx.nd.zeros_like(past_is_pad_agg),
            mx.nd.ones_like(past_is_pad_agg),
        )

        past_observed_values_agg = (
            utils.agg_series(
                past_observed_values.slice_axis(
                    axis=1, begin=-base_history_length, end=None
                ),
                agg_multiple=agg_multiple,
            )
            .squeeze(axis=-1)
            .slice_axis(
                axis=1, begin=-self.models[agg_level].history_length, end=None
            )
        )

        past_observed_values_agg = mx.nd.where(
            # We sum observed values of base time series at `agg_multiple` time steps;
            # if all of them are 1, then the observed value for the aggregated time series is 1 and 0 otherwise.
            # We could redefine agg_series to actually compute mean, but overloading that term might cause other
            # problems later.
            past_observed_values_agg == agg_multiple,
            mx.nd.ones_like(past_observed_values_agg),
            mx.nd.zeros_like(past_observed_values_agg),
        )

        target_related_feat_agg = {
            "past_target": past_target_agg,
            "past_is_pad": past_is_pad_agg,
            "past_observed_values": past_observed_values_agg,
        }
        if future_target is not None:
            future_target_agg = utils.agg_series(
                future_target, agg_multiple=agg_multiple
            ).squeeze(axis=-1)

            future_observed_values_agg = utils.agg_series(
                future_observed_values, agg_multiple=agg_multiple
            ).squeeze(axis=-1)

            future_observed_values_agg = mx.nd.where(
                future_observed_values_agg == agg_multiple,
                mx.nd.ones_like(future_observed_values_agg),
                mx.nd.zeros_like(future_observed_values_agg),
            )

            target_related_feat_agg.update(
                {
                    "future_target": future_target_agg,
                    "future_observed_values": future_observed_values_agg,
                }
            )

        return target_related_feat_agg

    def _embeddings_to_distr(
        self,
        F,
        embeddings_at_all_levels: Tensor,
        scales: List,
    ) -> Distribution:
        distr_output = self.models[0].distr_output
        distr_args_at_all_levels: Dict = {
            arg_name: [] for arg_name in distr_output.args_dim.keys()
        }
        scales_ls = []

        start_ix = 0
        for i, num_nodes in enumerate(
            self.temporal_hierarchy.num_nodes_per_level
        ):
            end_ix = start_ix + num_nodes
            distr_args = self.models[i].proj_distr_args(
                embeddings_at_all_levels[..., start_ix:end_ix, :]
            )

            for j, arg_ls in enumerate(distr_args_at_all_levels.values()):
                arg_ls.append(distr_args[j])

            scales_ls.append(scales[i].broadcast_like(distr_args[0]))

            start_ix = end_ix

        # Last dimension contains parameters at all time-levels and aggregation can be done on it.
        distr_args_at_all_levels = {
            arg_name: F.concat(*arg_ls, dim=-1)
            for arg_name, arg_ls in distr_args_at_all_levels.items()
        }

        scale_at_all_levels = F.concat(*scales_ls, dim=-1)

        distr_at_all_levels = distr_output.distribution(
            distr_args=distr_args_at_all_levels.values(),
            scale=scale_at_all_levels,
        )

        if isinstance(distr_at_all_levels, TransformedPiecewiseLinear):
            distr_at_all_levels = TransformedPiecewiseLinear(
                base_distribution=gluonts_fixes.PiecewiseLinearWithSampling(
                    gamma=distr_at_all_levels.base_distribution.gamma,
                    slopes=distr_at_all_levels.base_distribution.slopes,
                    knot_spacings=distr_at_all_levels.base_distribution.knot_spacings,
                ),
                transforms=distr_at_all_levels.transforms,
            )

        return distr_at_all_levels

    def _distr_to_samples(
        self,
        distr_at_all_levels: Distribution,
        num_samples: int,
    ):
        if num_samples == 1:
            samples_at_all_levels = distr_at_all_levels.sample(
                num_samples=num_samples, dtype=self.dtype
            )

            # get rid of the redundant axis introduced by `sample`.
            samples_at_all_levels = samples_at_all_levels.squeeze(axis=0)
        else:
            samples_at_all_levels = distr_at_all_levels.sample_rep(
                num_samples=num_samples, dtype=self.dtype
            )

        return samples_at_all_levels


class COPDeepARTrainingNetwork(COPNetwork):
    @validated()
    def __init__(
        self,
        num_batches_per_epoch: int,
        epochs: int,
        warmstart_epoch_frac: float,
        num_samples_for_loss: int = 200,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.warmstart_epoch_frac = warmstart_epoch_frac
        self.epochs = epochs
        self.num_batches_per_epoch = num_batches_per_epoch
        self.batch_no = 0
        self.num_samples_for_loss = num_samples_for_loss

    # noinspection PyMethodOverriding,PyPep8Naming
    def hybrid_forward(
        self,
        F,
        feat_static_cat: Tensor,
        feat_static_real: Tensor,
        past_time_feat: Tensor,
        past_target: Tensor,
        past_observed_values: Tensor,
        past_is_pad: Optional[Tensor],
        future_time_feat: Tensor,
        future_target: Tensor,
        future_observed_values: Tensor,
        agg_features_dict: Dict,
    ) -> Tensor:
        """
        Computes the loss for training COPDeepAR, all inputs tensors
        representing time series have NTC layout.

        Parameters
        ----------
        F
        feat_static_cat : (batch_size, num_features)
        feat_static_real : (batch_size, num_features)
        past_time_feat : (batch_size, history_length, num_features)
        past_target : (batch_size, history_length, *target_shape)
        past_observed_values : (batch_size, history_length, *target_shape, seq_len)
        future_time_feat : (batch_size, prediction_length, num_features)
        future_target : (batch_size, prediction_length, *target_shape)
        future_observed_values : (batch_size, prediction_length, *target_shape)
        agg_features_dict: Dictionary of features for aggregated levels

        Returns loss with shape (batch_size, context + prediction_length, 1)
        -------
        """

        embeddings_at_all_levels_ls = []
        target_at_all_levels_ls = []
        scale_ls = []

        for i, agg_multiple in enumerate(
            self.temporal_hierarchy.agg_multiples
        ):
            if agg_multiple != 1:
                past_time_feat_agg = agg_features_dict[f"level_{i}"][
                    "past_time_feat_agg"
                ]
                future_time_feat_agg = agg_features_dict[f"level_{i}"][
                    "future_time_feat_agg"
                ]
            else:
                past_time_feat_agg = past_time_feat
                future_time_feat_agg = future_time_feat

            target_related_feat_agg = (
                self.get_target_related_feat_at_agg_level(
                    agg_level=i,
                    past_target=past_target,
                    past_is_pad=past_is_pad,
                    past_observed_values=past_observed_values,
                    future_target=future_target,
                    future_observed_values=future_observed_values,
                )
            )

            rnn_outputs, _, scale, _, _ = self.models[i].unroll_encoder(
                F=F,
                feat_static_cat=feat_static_cat,
                feat_static_real=feat_static_real,
                past_time_feat=past_time_feat_agg,
                future_time_feat=future_time_feat_agg,
                **target_related_feat_agg,
            )
            scale_ls.append(scale.expand_dims(axis=-1))

            # put together target sequence
            # (batch_size, seq_len, *target_shape)
            target = F.concat(
                target_related_feat_agg["past_target"].slice_axis(
                    axis=1,
                    begin=self.models[i].history_length
                    - self.models[i].context_length,
                    end=None,
                ),
                target_related_feat_agg["future_target"],
                dim=1,
            )

            # We reconcile blocks/windows of time steps: e.g., if we have 28 values of daily data, then we
            # reconcile 4 windows where each window has a length of 7 if number of leaves in the hierarchy is 7.
            window_size = self.temporal_hierarchy.num_leaves // agg_multiple
            num_windows = (
                self.models[i].context_length
                + self.models[i].prediction_length
            ) // window_size

            embeddings_at_all_levels_ls.append(
                rnn_outputs.reshape(
                    (
                        rnn_outputs.shape[0],
                        num_windows,
                        -1,
                        rnn_outputs.shape[-1],
                    )
                )
            )

            target_at_all_levels_ls.append(
                target.reshape((target.shape[0], num_windows, -1))
            )

        # Last dimension contains embeddings at all time-levels and message passing/aggregation can be done on it.
        # Shape: (bs, num_windows, total_num_time_steps_of_hierarchy, embedding_dim)
        embeddings_at_all_levels = F.concat(
            *embeddings_at_all_levels_ls, dim=-2
        )

        if self.use_gnn:
            embeddings_at_all_levels = self.gnn(embeddings_at_all_levels)

        distr_at_all_levels = self._embeddings_to_distr(
            F,
            embeddings_at_all_levels,
            scale_ls,
        )

        target_at_all_levels = F.concat(*target_at_all_levels_ls, dim=-1)

        if self.loss_function == "nll":
            loss = distr_at_all_levels.loss(x=target_at_all_levels)

            # Determine which epoch we are currently in.
            self.batch_no += 1
            epoch_no = self.batch_no // self.num_batches_per_epoch + 1
            epoch_frac = epoch_no / self.epochs

            if epoch_frac > self.warmstart_epoch_frac:
                print(
                    f"epoch_frac: {epoch_frac}. Switching the loss function to CRPS"
                )
                self.loss_function = "crps_univariate"

        else:
            samples_at_all_levels = self._distr_to_samples(
                distr_at_all_levels,
                num_samples=self.num_samples_for_loss,
            )

            if self.do_reconciliation:
                reconciled_samples_at_all_levels = reconcile_samples(
                    reconciliation_mat=self.M,
                    samples=samples_at_all_levels,
                    non_negative=self.non_negative,
                )
            else:
                reconciled_samples_at_all_levels = samples_at_all_levels

            loss = (
                EmpiricalDistribution(
                    samples=reconciled_samples_at_all_levels, event_dim=1
                )
                .loss(x=target_at_all_levels)
                .expand_dims(axis=-1)
            )

        return loss


class COPDeepARPredictionNetwork(COPNetwork):
    @validated()
    def __init__(
        self,
        return_forecasts_at_all_levels: bool = False,
        num_parallel_samples: int = 100,
        **kwargs,
    ) -> None:
        super().__init__(prediction=True, **kwargs)
        self.return_forecasts_at_all_levels = return_forecasts_at_all_levels
        self.num_parallel_samples = num_parallel_samples

    def _decode_one_window(
        self,
        F,
        model: DeepARPredictionNetwork,
        window_size: int,
        offset: int,
        static_feat: Tensor,
        past_target: Tensor,
        time_feat: Tensor,
        scale: Tensor,
        begin_states: List,
    ) -> Tuple[Tensor, Tensor]:
        """
        Computes RNN outputs by unrolling the LSTM starting with a initial
        input and state.

        Parameters
        ----------
        static_feat : Tensor
            static features. Shape: (batch_size, num_static_features).
        past_target : Tensor
            target history. Shape: (batch_size, history_length).
        time_feat : Tensor
            time features. Shape: (batch_size, prediction_length,
            num_time_features).
            Note: They still need to be for all `prediction_length` time steps.
            This function will slice the features it needs.
        scale : Tensor
            tensor containing the scale of each element in the batch.
            Shape: (batch_size, 1, 1).
        begin_states : List
            list of initial states for the LSTM layers. The shape of each
            tensor of the list should be (batch_size, num_cells)
        Returns
        --------
        Tensor
            A tensor containing sampled paths.
            Shape: (batch_size, num_sample_paths, window_size).
        """

        rnn_outputs_ls = []

        # for each future time-units we draw new samples for this time-unit and
        # update the state
        for k in range(offset, offset + window_size):
            # (batch_size * num_samples, 1, *target_shape, num_lags)
            lags = model.get_lagged_subsequences(
                F=F,
                sequence=past_target,
                sequence_length=model.history_length + k,
                indices=model.shifted_lags,
                subsequences_length=1,
            )

            # (batch_size * num_samples, 1, *target_shape, num_lags)
            lags_scaled = F.broadcast_div(lags, scale.expand_dims(axis=-1))

            # from (batch_size * num_samples, 1, *target_shape, num_lags)
            # to (batch_size * num_samples, 1, prod(target_shape) * num_lags)
            input_lags = F.reshape(
                data=lags_scaled,
                shape=(-1, 1, prod(model.target_shape) * len(model.lags_seq)),
            )

            # (batch_size * num_samples, 1, prod(target_shape) * num_lags +
            # num_time_features + num_static_features)
            decoder_input = F.concat(
                input_lags,
                time_feat.slice_axis(axis=1, begin=k, end=k + 1),
                # observed_values.expand_dims(axis=1),
                static_feat,
                dim=-1,
            )

            # output shape: (batch_size * num_samples, 1, num_cells)
            # state shape: (batch_size * num_samples, num_cells)
            rnn_outputs, begin_states = model.rnn.unroll(
                inputs=decoder_input,
                length=1,
                begin_state=begin_states,
                layout="NTC",
                merge_outputs=True,
            )
            rnn_outputs_ls.append(rnn_outputs)

            distr_args = model.proj_distr_args(rnn_outputs)

            # compute likelihood of target given the predicted parameters
            distr = model.distr_output.distribution(distr_args, scale=scale)

            # (batch_size * num_samples, 1, *target_shape)

            new_samples = distr.sample(dtype=self.dtype)

            # (batch_size * num_samples, seq_len, *target_shape)
            past_target = F.concat(past_target, new_samples, dim=1)

        # (batch_size * num_samples, prediction_length, *target_shape)
        rnn_outputs = F.concat(*rnn_outputs_ls, dim=1)

        return rnn_outputs, begin_states

    def sampling_decoder(
        self,
        F,
        state_ls,
        scale_ls,
        static_feat_ls,
        past_target_ls,
        future_time_feat_agg_ls,
    ):
        num_windows = (
            self.prediction_length // self.temporal_hierarchy.num_leaves
        )
        num_nodes_per_level = self.temporal_hierarchy.num_nodes_per_level

        reconciled_samples_at_all_levels_ls = []
        for j in range(num_windows):
            embeddings_at_all_levels_ls = []
            for i, agg_multiple in enumerate(
                self.temporal_hierarchy.agg_multiples
            ):
                rnn_outputs, states = self._decode_one_window(
                    F=F,
                    model=self.models[i],
                    window_size=num_nodes_per_level[i],
                    offset=j * num_nodes_per_level[i],
                    past_target=past_target_ls[i],
                    time_feat=future_time_feat_agg_ls[i],
                    static_feat=static_feat_ls[i],
                    scale=scale_ls[i],
                    begin_states=state_ls[i],
                )

                state_ls[i] = states
                embeddings_at_all_levels_ls.append(
                    rnn_outputs.reshape(
                        (rnn_outputs.shape[0], -1, rnn_outputs.shape[-1])
                    )
                )

            # Last dimension contains embeddings at all time-levels and message passing/aggregation can be done on it.
            # Shape: (bs, total_num_time_steps_of_hierarchy, embedding_dim)
            embeddings_at_all_levels = F.concat(
                *embeddings_at_all_levels_ls, dim=-2
            )

            if self.use_gnn:
                embeddings_at_all_levels = self.gnn(embeddings_at_all_levels)

            distr_at_all_levels = self._embeddings_to_distr(
                F,
                embeddings_at_all_levels,
                scale_ls,
            )

            samples_at_all_levels = self._distr_to_samples(
                distr_at_all_levels,
                num_samples=1,
            )

            if self.do_reconciliation:
                reconciled_samples_at_all_levels = reconcile_samples(
                    reconciliation_mat=self.M,
                    samples=samples_at_all_levels,
                    non_negative=self.non_negative,
                )
            else:
                reconciled_samples_at_all_levels = samples_at_all_levels

            rec_err = coherency_error(
                S=self.temporal_hierarchy.agg_mat,
                samples=reconciled_samples_at_all_levels.asnumpy(),
            )
            print(f"Reconciliation error: {rec_err}")

            cumsum_nodes_per_level = np.cumsum([0] + num_nodes_per_level)
            for i in range(len(self.temporal_hierarchy.agg_multiples)):
                # (batch_size * num_samples, seq_len, *target_shape)
                reconciled_samples = (
                    reconciled_samples_at_all_levels.slice_axis(
                        axis=-1,
                        begin=cumsum_nodes_per_level[i],
                        end=cumsum_nodes_per_level[i + 1],
                    )
                )
                past_target_ls[i] = F.concat(
                    past_target_ls[i], reconciled_samples, dim=1
                )

            reconciled_samples_at_all_levels_ls.append(
                reconciled_samples_at_all_levels.reshape(
                    shape=(
                        -1,
                        self.num_parallel_samples,
                        reconciled_samples_at_all_levels.shape[-1],
                    )
                ).expand_dims(axis=-2)
            )

        reconciled_samples_at_all_levels = F.concat(
            *reconciled_samples_at_all_levels_ls, dim=-2
        )
        print(reconciled_samples_at_all_levels.shape)
        return reconciled_samples_at_all_levels

    # noinspection PyMethodOverriding,PyPep8Naming
    def hybrid_forward(
        self,
        F,
        feat_static_cat: Tensor,  # (batch_size, num_features)
        feat_static_real: Tensor,  # (batch_size, num_features)
        past_time_feat: Tensor,  # (batch_size, history_length, num_features)
        past_target: Tensor,  # (batch_size, history_length, *target_shape)
        past_observed_values: Tensor,  # (batch_size, history_length, *target_shape)
        future_time_feat: Tensor,  # (batch_size, prediction_length, num_features)
        past_is_pad: Tensor,
        agg_features_dict: Dict,
    ) -> Tensor:
        """
        Predicts samples, all tensors should have NTC layout.
        Parameters
        ----------
        F
        feat_static_cat : (batch_size, num_features)
        feat_static_real : (batch_size, num_features)
        past_time_feat : (batch_size, history_length, num_features)
        past_target : (batch_size, history_length, *target_shape)
        past_observed_values : (batch_size, history_length, *target_shape)
        future_time_feat : (batch_size, prediction_length, num_features)
        agg_features_dict: Dictionary of features for aggregated levels

        Returns
        -------
        Tensor
            Predicted samples
        """

        (
            state_ls,
            scale_ls,
            static_feat_ls,
            past_target_ls,
            future_time_feat_agg_ls,
        ) = ([], [], [], [], [])

        for i, agg_multiple in enumerate(
            self.temporal_hierarchy.agg_multiples
        ):
            if agg_multiple != 1:
                past_time_feat_agg = agg_features_dict[f"level_{i}"][
                    "past_time_feat_agg"
                ]
                future_time_feat_agg = agg_features_dict[f"level_{i}"][
                    "future_time_feat_agg"
                ]
            else:
                past_time_feat_agg = past_time_feat
                future_time_feat_agg = future_time_feat

            target_related_feat_agg = (
                self.get_target_related_feat_at_agg_level(
                    agg_level=i,
                    past_target=past_target,
                    past_is_pad=past_is_pad,
                    past_observed_values=past_observed_values,
                )
            )

            # unroll the decoder in "prediction mode", i.e. with past data only
            _, states, scale, static_feat, imputed_sequence = self.models[
                i
            ].unroll_encoder(
                F=F,
                feat_static_cat=feat_static_cat,
                feat_static_real=feat_static_real,
                past_time_feat=past_time_feat_agg,
                future_observed_values=None,
                future_time_feat=None,
                future_target=None,
                **target_related_feat_agg,
            )

            # blows-up the dimension of each tensor to batch_size *
            # self.num_parallel_samples for increasing parallelism
            repeated_past_target = imputed_sequence.repeat(
                repeats=self.num_parallel_samples, axis=0
            )
            repeated_states = [
                s.repeat(repeats=self.num_parallel_samples, axis=0)
                for s in states
            ]
            repeated_time_feat = future_time_feat_agg.repeat(
                repeats=self.num_parallel_samples, axis=0
            )
            repeated_static_feat = static_feat.repeat(
                repeats=self.num_parallel_samples, axis=0
            ).expand_dims(axis=1)
            repeated_scale = scale.repeat(
                repeats=self.num_parallel_samples, axis=0
            )

            state_ls.append(repeated_states)
            scale_ls.append(repeated_scale)
            static_feat_ls.append(repeated_static_feat)
            past_target_ls.append(repeated_past_target)
            future_time_feat_agg_ls.append(repeated_time_feat)

        reconciled_samples_at_all_levels = self.sampling_decoder(
            F,
            state_ls=state_ls,
            scale_ls=scale_ls,
            static_feat_ls=static_feat_ls,
            past_target_ls=past_target_ls,
            future_time_feat_agg_ls=future_time_feat_agg_ls,
        )

        if self.return_forecasts_at_all_levels:
            return reconciled_samples_at_all_levels
        else:
            reconciled_samples_at_bottom_level = (
                reconciled_samples_at_all_levels.slice_axis(
                    axis=-1,
                    begin=-self.temporal_hierarchy.num_leaves,
                    end=None,
                )
            )

            reconciled_samples_at_bottom_level = (
                reconciled_samples_at_bottom_level.reshape(
                    (
                        reconciled_samples_at_bottom_level.shape[0],
                        reconciled_samples_at_bottom_level.shape[1],
                        -1,
                    )
                )
            )

            return reconciled_samples_at_bottom_level



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import List, Optional, Tuple, Union

import numpy as np
import torch
import torch.nn as nn
from torch.distributions import Distribution

from src.gluonts.core.component import validated
from src.gluonts.torch.modules.distribution_output import DistributionOutput
from pts.model import weighted_average
from pts.modules import MeanScaler, NOPScaler, FeatureEmbedder


def prod(xs):
    p = 1
    for x in xs:
        p *= x
    return p


class DeepARNetwork(nn.Module):
    @validated()
    def __init__(
        self,
        input_size: int,
        num_layers: int,
        num_cells: int,
        cell_type: str,
        history_length: int,
        context_length: int,
        prediction_length: int,
        distr_output: DistributionOutput,
        dropout_rate: float,
        cardinality: List[int],
        embedding_dimension: List[int],
        lags_seq: List[int],
        scaling: bool = True,
        dtype: np.dtype = np.float32,
    ) -> None:
        super().__init__()
        self.num_layers = num_layers
        self.num_cells = num_cells
        self.cell_type = cell_type
        self.history_length = history_length
        self.context_length = context_length
        self.prediction_length = prediction_length
        self.dropout_rate = dropout_rate
        self.cardinality = cardinality
        self.embedding_dimension = embedding_dimension
        self.num_cat = len(cardinality)
        self.scaling = scaling
        self.dtype = dtype

        self.lags_seq = lags_seq

        self.distr_output = distr_output
        rnn = {"LSTM": nn.LSTM, "GRU": nn.GRU}[self.cell_type]
        self.rnn = rnn(
            input_size=input_size,
            hidden_size=num_cells,
            num_layers=num_layers,
            dropout=dropout_rate,
            batch_first=True,
        )

        self.target_shape = distr_output.event_shape

        self.proj_distr_args = distr_output.get_args_proj(num_cells)

        self.embedder = FeatureEmbedder(
            cardinalities=cardinality, embedding_dims=embedding_dimension
        )

        if scaling:
            self.scaler = MeanScaler(keepdim=True)
        else:
            self.scaler = NOPScaler(keepdim=True)

    @staticmethod
    def get_lagged_subsequences(
        sequence: torch.Tensor,
        sequence_length: int,
        indices: List[int],
        subsequences_length: int = 1,
    ) -> torch.Tensor:
        """
        Returns lagged subsequences of a given sequence.
        Parameters
        ----------
        sequence : Tensor
            the sequence from which lagged subsequences should be extracted.
            Shape: (N, T, C).
        sequence_length : int
            length of sequence in the T (time) dimension (axis = 1).
        indices : List[int]
            list of lag indices to be used.
        subsequences_length : int
            length of the subsequences to be extracted.
        Returns
        --------
        lagged : Tensor
            a tensor of shape (N, S, C, I), where S = subsequences_length and
            I = len(indices), containing lagged subsequences. Specifically,
            lagged[i, j, :, k] = sequence[i, -indices[k]-S+j, :].
        """
        assert max(indices) + subsequences_length <= sequence_length, (
            f"lags cannot go further than history length, found lag {max(indices)} "
            f"while history length is only {sequence_length}"
        )
        assert all(lag_index >= 0 for lag_index in indices)

        lagged_values = []
        for lag_index in indices:
            begin_index = -lag_index - subsequences_length
            end_index = -lag_index if lag_index > 0 else None
            lagged_values.append(sequence[:, begin_index:end_index, ...])
        return torch.stack(lagged_values, dim=-1)

    def unroll_encoder(
        self,
        feat_static_cat: torch.Tensor,  # (batch_size, num_features)
        feat_static_real: torch.Tensor,  # (batch_size, num_features)
        past_time_feat: torch.Tensor,  # (batch_size, history_length, num_features)
        past_target: torch.Tensor,  # (batch_size, history_length, *target_shape)
        past_observed_values: torch.Tensor,  # (batch_size, history_length, *target_shape)
        future_time_feat: Optional[
            torch.Tensor
        ] = None,  # (batch_size, prediction_length, num_features)
        future_target: Optional[
            torch.Tensor
        ] = None,  # (batch_size, prediction_length, *target_shape)
    ) -> Tuple[
        torch.Tensor, Union[torch.Tensor, List], torch.Tensor, torch.Tensor
    ]:
        if future_time_feat is None or future_target is None:
            time_feat = past_time_feat[
                :, self.history_length - self.context_length :, ...
            ]
            sequence = past_target
            sequence_length = self.history_length
            subsequences_length = self.context_length
        else:
            time_feat = torch.cat(
                (
                    past_time_feat[
                        :, self.history_length - self.context_length :, ...
                    ],
                    future_time_feat,
                ),
                dim=1,
            )
            sequence = torch.cat((past_target, future_target), dim=1)
            sequence_length = self.history_length + self.prediction_length
            subsequences_length = self.context_length + self.prediction_length

        lags = self.get_lagged_subsequences(
            sequence=sequence,
            sequence_length=sequence_length,
            indices=self.lags_seq,
            subsequences_length=subsequences_length,
        )

        # scale is computed on the context length last units of the past target
        # scale shape is (batch_size, 1, *target_shape)
        _, scale = self.scaler(
            past_target[:, -self.context_length :, ...],
            past_observed_values[:, -self.context_length :, ...],
        )

        # (batch_size, num_features)
        embedded_cat = self.embedder(feat_static_cat)

        # in addition to embedding features, use the log scale as it can help
        # prediction too
        # (batch_size, num_features + prod(target_shape))
        static_feat = torch.cat(
            (
                embedded_cat,
                feat_static_real,
                (
                    scale.log()
                    if len(self.target_shape) == 0
                    else scale.squeeze(1).log()
                ),
            ),
            dim=1,
        )

        # (batch_size, subsequences_length, num_features + 1)
        repeated_static_feat = static_feat.unsqueeze(1).expand(
            -1, subsequences_length, -1
        )

        # (batch_size, sub_seq_len, *target_shape, num_lags)
        lags_scaled = lags / scale.unsqueeze(-1)

        # from (batch_size, sub_seq_len, *target_shape, num_lags)
        # to (batch_size, sub_seq_len, prod(target_shape) * num_lags)
        input_lags = lags_scaled.reshape(
            (
                -1,
                subsequences_length,
                len(self.lags_seq) * prod(self.target_shape),
            )
        )

        # (batch_size, sub_seq_len, input_dim)
        inputs = torch.cat(
            (input_lags, time_feat, repeated_static_feat), dim=-1
        )

        # unroll encoder
        outputs, state = self.rnn(inputs)

        # outputs: (batch_size, seq_len, num_cells)
        # state: list of (num_layers, batch_size, num_cells) tensors
        # scale: (batch_size, 1, *target_shape)
        # static_feat: (batch_size, num_features + prod(target_shape))
        return outputs, state, scale, static_feat


class DeepARTrainingNetwork(DeepARNetwork):
    def distribution(
        self,
        feat_static_cat: torch.Tensor,
        feat_static_real: torch.Tensor,
        past_time_feat: torch.Tensor,
        past_target: torch.Tensor,
        past_observed_values: torch.Tensor,
        future_time_feat: torch.Tensor,
        future_target: torch.Tensor,
        future_observed_values: torch.Tensor,
    ) -> Distribution:
        rnn_outputs, _, scale, _ = self.unroll_encoder(
            feat_static_cat=feat_static_cat,
            feat_static_real=feat_static_real,
            past_time_feat=past_time_feat,
            past_target=past_target,
            past_observed_values=past_observed_values,
            future_time_feat=future_time_feat,
            future_target=future_target,
        )

        distr_args = self.proj_distr_args(rnn_outputs)

        return self.distr_output.distribution(distr_args, scale=scale)

    def forward(
        self,
        feat_static_cat: torch.Tensor,
        feat_static_real: torch.Tensor,
        past_time_feat: torch.Tensor,
        past_target: torch.Tensor,
        past_observed_values: torch.Tensor,
        future_time_feat: torch.Tensor,
        future_target: torch.Tensor,
        future_observed_values: torch.Tensor,
    ) -> torch.Tensor:
        distr = self.distribution(
            feat_static_cat=feat_static_cat,
            feat_static_real=feat_static_real,
            past_time_feat=past_time_feat,
            past_target=past_target,
            past_observed_values=past_observed_values,
            future_time_feat=future_time_feat,
            future_target=future_target,
            future_observed_values=future_observed_values,
        )

        # put together target sequence
        # (batch_size, seq_len, *target_shape)
        target = torch.cat(
            (
                past_target[
                    :, self.history_length - self.context_length :, ...
                ],
                future_target,
            ),
            dim=1,
        )

        # (batch_size, seq_len)
        loss = -distr.log_prob(target)

        # (batch_size, seq_len, *target_shape)
        observed_values = torch.cat(
            (
                past_observed_values[
                    :, self.history_length - self.context_length :, ...
                ],
                future_observed_values,
            ),
            dim=1,
        )

        # mask the loss at one time step iff one or more observations is missing in the target dimensions
        # (batch_size, seq_len)
        loss_weights = (
            observed_values
            if (len(self.target_shape) == 0)
            else observed_values.min(dim=-1, keepdim=False)
        )

        weighted_loss = weighted_average(loss, weights=loss_weights)

        return weighted_loss, loss


class DeepARPredictionNetwork(DeepARNetwork):
    def __init__(self, num_parallel_samples: int = 100, **kwargs) -> None:
        super().__init__(**kwargs)
        self.num_parallel_samples = num_parallel_samples

        # for decoding the lags are shifted by one, at the first time-step
        # of the decoder a lag of one corresponds to the last target value
        self.shifted_lags = [l - 1 for l in self.lags_seq]

    def sampling_decoder(
        self,
        static_feat: torch.Tensor,
        past_target: torch.Tensor,
        time_feat: torch.Tensor,
        scale: torch.Tensor,
        begin_states: Union[torch.Tensor, List[torch.Tensor]],
    ) -> torch.Tensor:
        """
        Computes sample paths by unrolling the RNN starting with a initial
        input and state.

        Parameters
        ----------
        static_feat : Tensor
            static features. Shape: (batch_size, num_static_features).
        past_target : Tensor
            target history. Shape: (batch_size, history_length).
        time_feat : Tensor
            time features. Shape: (batch_size, prediction_length, num_time_features).
        scale : Tensor
            tensor containing the scale of each element in the batch. Shape: (batch_size, 1, 1).
        begin_states : List or Tensor
            list of initial states for the LSTM layers or tensor for GRU.
            the shape of each tensor of the list should be (num_layers, batch_size, num_cells)
        Returns
        --------
        Tensor
            A tensor containing sampled paths.
            Shape: (batch_size, num_sample_paths, prediction_length).
        """

        # blows-up the dimension of each tensor to batch_size * self.num_parallel_samples for increasing parallelism
        repeated_past_target = past_target.repeat_interleave(
            repeats=self.num_parallel_samples, dim=0
        )
        repeated_time_feat = time_feat.repeat_interleave(
            repeats=self.num_parallel_samples, dim=0
        )
        repeated_static_feat = static_feat.repeat_interleave(
            repeats=self.num_parallel_samples, dim=0
        ).unsqueeze(1)
        repeated_scale = scale.repeat_interleave(
            repeats=self.num_parallel_samples, dim=0
        )
        if self.cell_type == "LSTM":
            repeated_states = [
                s.repeat_interleave(repeats=self.num_parallel_samples, dim=1)
                for s in begin_states
            ]
        else:
            repeated_states = begin_states.repeat_interleave(
                repeats=self.num_parallel_samples, dim=1
            )

        future_samples = []

        # for each future time-units we draw new samples for this time-unit and update the state
        for k in range(self.prediction_length):
            # (batch_size * num_samples, 1, *target_shape, num_lags)
            lags = self.get_lagged_subsequences(
                sequence=repeated_past_target,
                sequence_length=self.history_length + k,
                indices=self.shifted_lags,
                subsequences_length=1,
            )

            # (batch_size * num_samples, 1, *target_shape, num_lags)
            lags_scaled = lags / repeated_scale.unsqueeze(-1)

            # from (batch_size * num_samples, 1, *target_shape, num_lags)
            # to (batch_size * num_samples, 1, prod(target_shape) * num_lags)
            input_lags = lags_scaled.reshape(
                (-1, 1, prod(self.target_shape) * len(self.lags_seq))
            )

            # (batch_size * num_samples, 1, prod(target_shape) * num_lags + num_time_features + num_static_features)
            decoder_input = torch.cat(
                (
                    input_lags,
                    repeated_time_feat[:, k : k + 1, :],
                    repeated_static_feat,
                ),
                dim=-1,
            )

            # output shape: (batch_size * num_samples, 1, num_cells)
            # state shape: (batch_size * num_samples, num_cells)
            rnn_outputs, repeated_states = self.rnn(
                decoder_input, repeated_states
            )

            distr_args = self.proj_distr_args(rnn_outputs)

            # compute likelihood of target given the predicted parameters
            distr = self.distr_output.distribution(
                distr_args, scale=repeated_scale
            )

            # (batch_size * num_samples, 1, *target_shape)
            new_samples = distr.sample()

            # (batch_size * num_samples, seq_len, *target_shape)
            repeated_past_target = torch.cat(
                (repeated_past_target, new_samples), dim=1
            )
            future_samples.append(new_samples)

        # (batch_size * num_samples, prediction_length, *target_shape)
        samples = torch.cat(future_samples, dim=1)

        # (batch_size, num_samples, prediction_length, *target_shape)
        return samples.reshape(
            (
                (-1, self.num_parallel_samples)
                + (self.prediction_length,)
                + self.target_shape
            )
        )

    # noinspection PyMethodOverriding,PyPep8Naming
    def forward(
        self,
        feat_static_cat: torch.Tensor,  # (batch_size, num_features)
        feat_static_real: torch.Tensor,  # (batch_size, num_features)
        past_time_feat: torch.Tensor,  # (batch_size, history_length, num_features)
        past_target: torch.Tensor,  # (batch_size, history_length, *target_shape)
        past_observed_values: torch.Tensor,  # (batch_size, history_length, *target_shape)
        future_time_feat: torch.Tensor,  # (batch_size, prediction_length, num_features)
    ) -> torch.Tensor:
        """
        Predicts samples, all tensors should have NTC layout.
        Parameters
        ----------
        feat_static_cat : (batch_size, num_features)
        feat_static_real : (batch_size, num_features)
        past_time_feat : (batch_size, history_length, num_features)
        past_target : (batch_size, history_length, *target_shape)
        past_observed_values : (batch_size, history_length, *target_shape)
        future_time_feat : (batch_size, prediction_length, num_features)

        Returns
        -------
        Tensor
            Predicted samples
        """

        # unroll the decoder in "prediction mode", i.e. with past data only
        _, state, scale, static_feat = self.unroll_encoder(
            feat_static_cat=feat_static_cat,
            feat_static_real=feat_static_real,
            past_time_feat=past_time_feat,
            past_target=past_target,
            past_observed_values=past_observed_values,
            future_time_feat=None,
            future_target=None,
        )

        return self.sampling_decoder(
            past_target=past_target,
            time_feat=future_time_feat,
            static_feat=static_feat,
            scale=scale,
            begin_states=state,
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.


from typing import Optional, List
from collections import OrderedDict
import warnings

import torch as pt
from torch import Tensor, BoolTensor
from torch import nn

from .attention import (
    SelfAttention,
    GroupSelfAttention,
    CompositeAttention,
    InterAttention,
)
from .activations import Swish

add_weightnorm = False
if add_weightnorm:
    from .weightnorm import Linear, Conv1d
else:
    Linear = nn.Linear
    Conv1d = nn.Conv1d


class PositionwiseFFN(nn.Module):
    """
    Positionwise feedforward network in transformers.
    """

    def __init__(
        self,
        d_model: int,
        d_hidden: int,
        pre_ln: bool = False,
        dropout: float = 0.0,
    ):
        super(PositionwiseFFN, self).__init__()
        self.linear1 = Linear(d_model, d_hidden)
        self.activation = nn.GELU()
        self.dropout = nn.Dropout(dropout)
        self.linear2 = Linear(d_hidden, d_model)
        self.layer_norm = nn.LayerNorm(d_model)
        self.pre_ln = pre_ln

    def forward(self, x: Tensor) -> Tensor:
        if self.pre_ln:
            y = self.layer_norm(x)
        else:
            y = x
        y = self.linear1(y)
        y = self.activation(y)
        y = self.dropout(y)
        y = self.linear2(y)
        y = self.dropout(y)
        y = y + x
        if not self.pre_ln:
            y = self.layer_norm(y)
        return y


class TransformerEncoderBlock(nn.Module):
    def __init__(
        self,
        d_qk: int,
        d_value: int,
        d_ffn: int,
        n_head: int,
        n_groups: int = 1,
        bidirectional: bool = False,
        apply_rel_dist: Optional[str] = None,
        composite_attention: bool = False,
        pre_ln: bool = False,
        dropout: float = 0.0,
        temperature: float = 1.0,
    ):
        super(TransformerEncoderBlock, self).__init__()
        if n_groups < 1:
            raise ValueError(
                f"n_groups must be a positive integer, but {n_groups} is given"
            )
        elif n_groups == 1:
            if d_qk != d_value:
                warning.warn(
                    "When n_groups=1, `d_qk` and `d_value` is suggested to be equal, "
                    f"but are {d_qk} and {d_value}, respectively. "
                    "d_qk will be ignored."
                )
            self.shape_qk = self.composite_attention = False
            self.attention = SelfAttention(
                d_hidden=d_value,
                n_head=n_head,
                bidirectional=bidirectional,
                apply_rel_dist=apply_rel_dist,
                dropout=dropout,
                temperature=temperature,
            )
        else:
            self.shape_qk = True
            self.composite_attention = composite_attention
            Attention = (
                CompositeAttention
                if composite_attention
                else GroupSelfAttention
            )
            self.attention = Attention(
                d_qk=d_qk,
                d_value=d_value,
                n_head=n_head,
                n_groups=n_groups,
                bidirectional=bidirectional,
                apply_rel_dist=apply_rel_dist,
                dropout=dropout,
                temperature=temperature,
            )
        self.pre_ln = pre_ln
        self.layer_norm = nn.LayerNorm(d_value)
        self.dropout = nn.Dropout(dropout)
        self.ffn = PositionwiseFFN(
            d_model=d_value, d_hidden=d_ffn, dropout=dropout, pre_ln=pre_ln
        )

    def forward(
        self,
        value: Tensor,
        shape: Optional[pt.Tensor] = None,
        estimate: Optional[pt.Tensor] = None,
        mask: Optional[BoolTensor] = None,
    ) -> Tensor:
        if self.shape_qk and shape is None:
            raise ValueError("A separated shape input is required.")
        if self.composite_attention and estimate is None:
            raise ValueError("A separated estimate input is required")
        if self.pre_ln:
            x = self.layer_norm(value)
        else:
            x = value
        if not self.shape_qk:
            x = self.attention(x, mask=mask)
        else:
            if shape is None:
                raise ValueError(
                    "if shape attention is enabled, "
                    "a shape tensor must be provided."
                )
            if not self.composite_attention:
                x = self.attention(x, shape, mask=mask)
            else:
                if estimate is None:
                    raise ValueError(
                        "if composite attention is enabled, "
                        "an estimate tensor must be provided."
                    )
                x = self.attention(x, shape, estimate, mask=mask)
        x = self.dropout(x)
        value = value + x
        if not self.pre_ln:
            value = self.layer_norm(value)
        value = self.ffn(value)
        return value


class TransformerDecoderBlock(nn.Module):
    def __init__(
        self,
        d_query: int,
        d_src: int,
        d_ffn: int,
        n_head: int,
        bidirectional: bool = False,
        apply_rel_dist: Optional[str] = None,
        pre_ln: bool = False,
        dropout: float = 0.0,
        temperature: float = 1.0,
    ):
        super(TransformerDecoderBlock, self).__init__()
        self.pre_ln = pre_ln
        self.self_attention = SelfAttention(
            d_query,
            n_head,
            bidirectional=bidirectional,
            apply_rel_dist=apply_rel_dist,
            dropout=dropout,
            temperature=temperature,
        )
        if pre_ln:
            self.self_attn_layer_norm = nn.LayerNorm(d_query)
        else:
            self.self_attn_layer_norm = nn.LayerNorm(d_query)
        self.inter_attention = InterAttention(
            d_query,
            d_src,
            n_head,
            apply_kv_proj=True,
            apply_rel_dist=apply_rel_dist,
            dropout=dropout,
        )
        self.inter_attn_layer_norm = nn.LayerNorm(d_query)
        self.ffn = PositionwiseFFN(
            d_model=d_query,
            d_hidden=d_ffn,
            dropout=dropout,
            pre_ln=pre_ln,
        )
        self.dropout = nn.Dropout(dropout)

    def forward(
        self,
        x: Tensor,
        src: Tensor,
        mask: Optional[BoolTensor] = None,
        src_mask: Optional[BoolTensor] = None,
    ):
        if self.pre_ln:
            y = self.self_attn_layer_norm(x)
        else:
            y = x
        y = self.self_attention(y, mask=mask)
        y = self.dropout(y)
        x = y + x
        if not self.pre_ln:
            x = self.self_attn_layer_norm(x)
            y = x
        else:
            y = self.inter_attn_layer_norm(x)
        y = self.inter_attention(y, src, src_mask=src_mask)
        y = self.dropout(y)
        x = y + x
        if not self.pre_ln:
            x = self.inter_attn_layer_norm(x)
        x = self.ffn(x)
        return x



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from dataclasses import asdict, dataclass, MISSING
from pathlib import Path
from typing import Any, Dict, List, Optional, TypeVar
from src.gluonts.model.estimator import Estimator
from src.gluonts.model.predictor import Predictor
from mxnet.gluon import nn
from tsbench.config.dataset import DatasetConfig
from tsbench.gluonts import TimedTrainer
from tsbench.src.gluonts.callbacks import Callback


@dataclass(frozen=True)  # frozen=True is required to make this type hashable
class ModelConfig:
    """
    A model configuration describes the configuration of a GluonTS model and
    allows instantiating an estimator from it.

    Compared to an estimator, this configuration class can easily be hashed and
    different objects with the same configuration are considered equal under
    the `==` operator.
    """

    @classmethod
    def name(cls) -> str:
        """
        Returns the canonical name for the model described by the
        configuration.
        """
        raise NotImplementedError

    @classmethod
    def hyperparameters(cls) -> Dict[str, bool]:
        """
        Returns the list of hyperparameters that are expected by this model
        configuration.

        For each hyperparameter, it provides a boolean whether there exists a
        default value.
        """
        # pylint: disable=no-member
        return {
            key: field.default is not MISSING
            for key, field in cls.__dataclass_fields__.items()  # type: ignore
        }

    def save_predictor(self, predictor: Predictor, path: Path) -> None:
        """
        Saves the predictor associated with the model configuration to the
        specified path. By default, this simply serializes the predictor.

        Args:
            predictor: The predictor to save.
            path: The directory where to save the predictor.
        """
        predictor.serialize(path)

    def load_predictor(self, path: Path) -> Predictor:
        """
        Loads the predictor from the specified path.

        Args:
            path: The directory from which to load the predictor.

        Returns:
            The predictor which was loaded.
        """
        return Predictor.deserialize(path)

    @property
    def prediction_samples(self) -> int:
        """
        The number of samples which should be produced when sampling during
        inference.
        """
        return 100

    @property
    def prefers_parallel_predictions(self) -> bool:
        """
        Returns whether predictions should be generated in parallel.
        """
        return False

    def max_time_series_length(self, _config: DatasetConfig) -> Optional[int]:
        """
        The maximum length a time series may have to be used for prediction. If
        an integer N is provided, only the most recent N observations are used.

        Args:
            config: The dataset for which to determine the maximum time series length.
        """
        return None

    def create_estimator(
        self,
        freq: str,
        prediction_length: int,
        time_features: bool,
        training_time: float,
        validation_milestones: List[float],
        callbacks: List[Callback],
    ) -> Estimator:
        """
        Creates a GluonTS estimator from the model's configuration.

        Args:
            freq: The frequency of the data that the estimator is created for.
            prediction_length: The required length of the model predictions.
            training_time: The number of seconds to train for.
            validation_milestones: The milestones used for running validation.
            callbacks: An optional list of callbacks to use during training.

        Returns:
            An initialized GluonTS estimator.
        """
        raise NotImplementedError

    def asdict(self) -> Dict[str, Any]:
        """
        Returns a dictionary representation of the model configuration.
        """
        return {
            "model": self.name(),
            **{
                (
                    k
                    if isinstance(self, TrainConfig)
                    and k in TrainConfig.training_hyperparameters()
                    else f"{self.name()}_{k}"
                ): v
                for k, v in asdict(self).items()
            },
        }


# -------------------------------------------------------------------------------------------------
C = TypeVar("C", bound="TrainConfig")


@dataclass(frozen=True)
class TrainConfig:
    """
    The training configuration may be derived by any estimator which is trained
    via the GluonTS `Trainer` class.
    """

    @classmethod
    def training_hyperparameters(cls) -> List[str]:
        """
        Returns the list of hyperparameters that are used for trainable models.

        Returns:
            The list of training hyperparameters.
        """
        # pylint: disable=no-member
        return list(cls.__dataclass_fields__.keys())  # type: ignore

    training_fraction: float = 1.0
    learning_rate: float = 1e-3
    context_length_multiple: int = 1

    def create_predictor(
        self, estimator: Estimator, network: nn.HybridBlock
    ) -> Predictor:
        """
        Creates a predictor from the provided network. This method is required
        to be implemented by every model configuration which describes a
        trainable model.

        Args:
            estimator: The estimator for which to obtain the predictor. Must have been obtained
            from the same model configuration. Otherwise, behavior is undefined.
            network: The network which has been trained.

        Returns:
            The predictor created from the network.
        """
        raise NotImplementedError

    def _create_trainer(
        self,
        training_time: float,
        validation_milestones: List[float],
        callbacks: List[Callback],
    ) -> TimedTrainer:
        return TimedTrainer(
            training_time=training_time,
            validation_milestones=validation_milestones,
            learning_rate=self.learning_rate,
            callbacks=callbacks,
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import torch
import torch.nn as nn
import torch.optim as optim
from tqdm.auto import tqdm
from typing import List
from sparse_layer import SparseNet


def forward_model(model, inputs):
    samples = model(**inputs)
    sample_means = torch.mean(samples, axis=1)
    return samples, sample_means


def generate_adv_target(mode, future_target, factor=1.5):
    if mode == "over":
        adv_target = factor * future_target
    elif mode == "under":
        adv_target = factor * future_target
    elif mode == "zero":
        adv_target = torch.zeros_like(future_target)
    else:
        raise Exception("No such mode")
    return adv_target


class SparseLayerAttack:
    def __init__(self, model, params, input_names):
        self.model = model
        self.params = params
        self.input_names = input_names
        self.model.dropout_rate = 0
        self.model.rnn.dropout = 0
        self.model.train()

    def attack_batch(self, batch, true_future_target):
        res = {}
        batch_size, context_length, dim = batch["past_target_cdf"].shape
        max_norm = self.params.max_norm

        with torch.no_grad():
            inputs = {key: batch[key] for key in self.input_names}
            _, sample_mu = forward_model(self.model, inputs)
            future_target = sample_mu
            loss_function = nn.MSELoss(reduction="none")
            best_loss = [0] * len(self.params.sparsity)

        for mode in self.params.modes:
            print("Mode ", mode)

            # generate target
            adv_target = generate_adv_target(
                mode, future_target, self.params.factor
            )

            # traverse over each sparsity
            for i, s in enumerate(self.params.sparsity):
                # Generate sparse attack using sparse layer
                sparse_layer = SparseNet(
                    context_length=context_length,
                    target_dim=dim,
                    target_item=self.params.target_items[0],
                    hidden_dim=40,
                    m=s,
                    max_norm=max_norm,
                ).cuda()
                optimizer = optim.Adam(
                    sparse_layer.parameters(), lr=self.params.learning_rate
                )

                # iteration steps
                for _ in tqdm(range(self.params.n_iterations)):
                    optimizer.zero_grad()
                    sparse_delta = sparse_layer(
                        batch["past_target_cdf"], n_sample=100
                    ).view(batch_size, context_length, dim)
                    perturbed_inputs = dict(
                        [(key, batch[key]) for key in self.input_names]
                    )
                    perturbed_inputs["past_target_cdf"] = batch[
                        "past_target_cdf"
                    ] * (1 + sparse_delta)
                    _, perturbed_mu = forward_model(
                        self.model, perturbed_inputs
                    )
                    loss = loss_function(perturbed_mu, adv_target)[
                        :, self.params.attack_idx, self.params.target_items
                    ].mean()
                    loss.backward()
                    optimizer.step()

                with torch.no_grad():
                    # construct perturbed inputs
                    perturbation = sparse_layer(
                        batch["past_target_cdf"], n_sample=100
                    ).view(batch_size, context_length, dim)
                    perturbed_inputs = {
                        key: batch[key] for key in self.input_names
                    }
                    perturbed_inputs["past_target_cdf"] = perturbed_inputs[
                        "past_target_cdf"
                    ] * (1 + perturbation)

                    # get model outputs
                    _, perturbed_mu = forward_model(
                        self.model, perturbed_inputs
                    )
                    # compute loss
                    loss = loss_function(
                        perturbed_mu[
                            :, self.params.attack_idx, self.params.target_items
                        ],
                        true_future_target[
                            :, self.params.attack_idx, self.params.target_items
                        ],
                    ).mean()
                    if loss >= best_loss[i]:
                        best_loss[i] = loss
                        res[s] = perturbation.data

        return res


class AttackLoss(nn.Module):
    def __init__(
        self,
        c: float,
        attack_idx: List[int],
        target_items: List[int],
        device="cuda" if torch.cuda.is_available() else "cpu",
    ):
        super(AttackLoss, self).__init__()
        self.c = c
        self.attack_idx = attack_idx
        self.device = device
        self.target_items = target_items

    def forward(self, perturbation, mu, adv_target):
        # mu: batch x prediction_length x dim
        mu = mu[:, self.attack_idx, :]  # batch x 1 x dim

        adv_target = adv_target[:, self.attack_idx, :]  # batch x 1 x dim

        loss_function = nn.MSELoss(reduction="none")

        distance_tensor = loss_function(adv_target, mu)[
            ..., self.target_items
        ].sum(
            -1
        )  # batch x len(attack_idx)
        distance_per_sample = distance_tensor.sum(axis=1)  # batch
        distance = distance_per_sample.sum()  # scalar

        zero = torch.zeros(perturbation.shape).to(self.device)
        norm_tensor = loss_function(perturbation, zero).sum(
            -1
        )  # batch x context_length
        norm_per_sample = norm_tensor.sum(axis=1)  # batch
        norm = norm_per_sample.sum()  # scalar

        loss_per_sample = distance_per_sample
        loss = loss_per_sample.sum()

        return (
            norm_per_sample,
            distance_per_sample,
            loss_per_sample,
            norm,
            distance,
            loss,
        )


class AttackModule(nn.Module):
    def __init__(self, model, params, c, batch, input_names):
        super(AttackModule, self).__init__()

        self.model = model
        self.params = params
        self.c = c
        self.batch = batch
        self.input_names = input_names

        self.attack_loss = AttackLoss(
            c,
            attack_idx=self.params.attack_idx,
            device=self.params.device,
            target_items=self.params.target_items,
        )

        # Initialize perturbation: batch x context length x target dim
        self.perturbation = nn.Parameter(
            torch.zeros_like(batch["past_target_cdf"]), requires_grad=True
        )
        self.perturbation.to(self.params.device)

    # Returns predicted mean and scale
    def forward(self, num_parallel_samples=100):
        perturbed_inputs = dict(
            [(key, self.batch[key]) for key in self.input_names]
        )
        perturbed_inputs["past_target_cdf"] = self.batch["past_target_cdf"] * (
            1 + self.perturbation
        )
        _, samples = forward_model(self.model, perturbed_inputs)
        return samples

    def get_perturbation(self):
        return self.perturbation

    def get_grad(self):
        return self.perturbation.grad


class SparseAttackModule(nn.Module):
    def __init__(self, model, params, c, batch, input_names, attack_item):
        super(SparseAttackModule, self).__init__()

        self.model = model
        self.params = params
        self.c = c
        self.batch = batch
        self.input_names = input_names
        self.attack_item = attack_item

        self.attack_loss = AttackLoss(
            c,
            attack_idx=self.params.attack_idx,
            device=self.params.device,
            target_items=self.params.target_items,
        )

        # Initialize perturbation: batch x context length x target dim
        self.perturbation = nn.Parameter(
            torch.zeros_like(batch["past_target_cdf"]), requires_grad=True
        )
        self.perturbation.to(self.params.device)
        self.mask = torch.zeros_like(
            self.perturbation, device=self.params.device
        )
        self.mask[..., attack_item] = 1

    # Returns predicted mean and scale
    def forward(self, num_parallel_samples=100):
        perturbed_inputs = dict(
            [(key, self.batch[key]) for key in self.input_names]
        )
        perturbed_inputs["past_target_cdf"] = self.batch["past_target_cdf"] * (
            1 + self.perturbation * self.mask
        )
        _, samples = forward_model(self.model, perturbed_inputs)
        return samples

    def get_perturbation(self):
        return self.perturbation * self.mask

    def get_grad(self):
        return self.perturbation.grad * self.mask


class Attack:
    def __init__(self, model, params, input_names):
        self.model = model
        self.params = params
        self.input_names = input_names
        self.model.dropout_rate = 0
        self.model.rnn.dropout = 0
        self.model.train()
        self.max_norm = params.max_norm

    def attack_step(
        self, attack_module, optimizer, adv_target, ground_truth=None
    ):
        optimizer.zero_grad()

        perturbed_mu = attack_module()
        batch_size = perturbed_mu.shape[0]
        max_norm = self.max_norm

        # targeted loss
        _, _, _, _, _, loss = attack_module.attack_loss(
            attack_module.get_perturbation(), perturbed_mu, adv_target
        )

        # backward pass
        loss.backward()

        # renorm grad
        grad_norm = (
            attack_module.get_grad().view(batch_size, -1).norm(p=2, dim=1)
        )
        attack_module.perturbation.grad.div_(grad_norm.view(batch_size, 1, 1))

        # back-propogation
        optimizer.step()
        attack_module.perturbation.data.clamp_(min=-max_norm, max=max_norm)

    def attack_batch(self, batch, true_future_target):
        res = {}

        # generate adv target
        with torch.no_grad():
            inputs = {key: batch[key] for key in self.input_names}
            _, sample_mu = forward_model(self.model, inputs)
            future_target = sample_mu
            attack_idx = self.params.attack_idx
            target_items = self.params.target_items
            loss_function = nn.MSELoss(reduction="none")
            best_dense = 0

        for mode in self.params.modes:
            print("Mode: ", mode)

            # Generate dense attack
            attack_module = AttackModule(
                model=self.model,
                params=self.params,
                c=0,
                batch=batch,
                input_names=self.input_names,
            )
            adv_target = generate_adv_target(
                future_target=future_target,
                mode=mode,
                factor=self.params.factor,
            )

            optimizer = optim.Adam(
                [attack_module.perturbation], lr=self.params.learning_rate
            )

            for _ in tqdm(range(self.params.n_iterations)):
                self.attack_step(attack_module, optimizer, adv_target)

            # evaluate attack
            with torch.no_grad():
                perturbed_mu = attack_module()
                loss = loss_function(
                    true_future_target[:, attack_idx, target_items],
                    perturbed_mu[:, attack_idx, target_items],
                ).mean()
                if loss >= best_dense:
                    best_dense = loss
                    res["dense"] = attack_module.get_perturbation().data

        return res



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import List, Optional
from ._base import CandidateGenerator, T


class ReplayCandidateGenerator(CandidateGenerator[T]):
    """
    The replay candidate generator simply returns the model configurations seen
    during training.

    If candidates are provided, they are returned as is.
    """

    def __init__(self) -> None:
        """
        Initializes a new replay candidate generator.
        """
        self.cache: List[T] = []

    def fit(self, configs: List[T]) -> None:
        self.cache = configs

    def generate(self, candidates: Optional[List[T]] = None) -> List[T]:
        # Assert trained
        assert self.cache, "Replay candidate generator has not been fitted."

        # Return candidates or cache
        return candidates or self.cache



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.


from typing import Tuple, Optional
import math

import torch as pt
from torch import Tensor, BoolTensor
from torch import nn
from torch.nn import init
from torch.nn.functional import linear, conv1d

from .base import Attention
from .posemb import SinusoidalPositionalEmbedding


class SelfAttention(Attention):
    """
    Self-attention module with q,k,v from the same input.

    Parameters
    ----------
    d_hidden : int
        hidden dimension
    n_head : int, optional
        number of attention heads, by default 1
    bias : bool, optional
        add bias term in input and output projections, by default True
    bidirectional : bool, optional
        if False, add a mask to avoid backward attention, by default False
    apply_rel_dist : Optional[str], optional
        add relative distance embeddings to dot-product attention, can be
            'add' (linearly combine key and dist),
            'dot' (dot product between key and dist),
            or None (disabled),
        by default None
    share_values : bool, optional
        if True, a value reprensentation is shared by all attention heads, by default False
        ref. https://arxiv.org/abs/1912.09363
    dropout : float, optional
        dropout rate, by default 0.0
    temperature : float, optional
        softmax temperature, by default 1.0
    """

    def __init__(
        self,
        d_hidden: int,
        n_head: int = 1,
        bias: bool = True,
        bidirectional: bool = False,
        apply_rel_dist: Optional[str] = None,
        share_values: bool = False,
        dropout: float = 0.0,
        temperature: float = 1.0,
    ):
        super(SelfAttention, self).__init__(
            d_hidden=d_hidden,
            n_head=n_head,
            bias=bias,
            dropout=dropout,
            temperature=temperature,
        )
        self.bidirectional = bidirectional
        self.apply_rel_dist = apply_rel_dist
        self.share_values = share_values

        d_value = self.d_head if self.share_values else d_hidden
        self._qk_proj_weight = nn.Parameter(Tensor(2 * d_hidden, d_hidden))
        self._value_proj_weight = nn.Parameter(Tensor(d_value, d_hidden))
        self._out_proj_weight = nn.Parameter(Tensor(d_hidden, d_hidden))
        if bias:
            self._qk_proj_bias = nn.Parameter(Tensor(2 * d_hidden))
            self._value_proj_bias = nn.Parameter(Tensor(d_value))
            self._out_proj_bias = nn.Parameter(Tensor(d_hidden))
        else:
            self.register_parameter("_in_proj_bias", None)
            self.register_parameter("_out_proj_bias", None)

        if self.apply_rel_dist is None:
            self.posemb = None
            self.register_parameter("_pos_proj_weight", None)
        else:
            assert self.apply_rel_dist in ["dot", "add"]
            self.posemb = SinusoidalPositionalEmbedding(d_hidden)
            self._pos_proj_weight = nn.Parameter(Tensor(d_hidden, d_hidden))
            if self.apply_rel_dist == "add":
                self._content_bias_weight = nn.Parameter(
                    Tensor(n_head, 1, self.d_head)
                )
                self._position_bias_weight = nn.Parameter(
                    Tensor(n_head, 1, self.d_head)
                )
            else:
                self.register_parameter("_content_bias_weight", None)
                self.register_parameter("_position_bias_weight", None)
        self._reset_parameters()

    def _reset_parameters(self):
        init.xavier_uniform_(self._qk_proj_weight)
        init.xavier_uniform_(self._value_proj_weight)
        init.xavier_uniform_(self._out_proj_weight)
        if self.bias:
            init.zeros_(self._qk_proj_bias)
            init.zeros_(self._value_proj_bias)
            init.zeros_(self._out_proj_bias)
        if self.apply_rel_dist is not None:
            init.xavier_uniform_(self._pos_proj_weight)
            if self.apply_rel_dist == "add":
                init.xavier_uniform_(self._content_bias_weight)
                init.xavier_uniform_(self._position_bias_weight)

    def _compute_qkv(self, x: Tensor) -> Tuple[Tensor, Tensor, Tensor]:
        qk = linear(x, self._qk_proj_weight, self._qk_proj_bias)
        q, k = qk.chunk(2, dim=-1)
        q = self._split_head(q)
        k = self._split_head(k)
        v = linear(x, self._value_proj_weight, self._value_proj_bias)
        if self.share_values:
            v = pt.stack([v] * self.n_head, dim=1)
        else:
            v = self._split_head(v)
        return q, k, v

    def _apply_mask(
        self, score: Tensor, key_mask: Optional[BoolTensor]
    ) -> Tensor:
        if not self.bidirectional:
            unid_mask = score.new_ones(*score.shape[-2:], dtype=pt.bool).triu(
                diagonal=1
            )
            score = score.masked_fill(unid_mask, float("-inf"))
        if key_mask is not None:
            key_mask = key_mask.unsqueeze(dim=1)  # head
            key_mask = key_mask.unsqueeze(dim=2)  # query
            score = score.masked_fill(key_mask, float("-inf"))
        return score

    def _compute_attn_score(
        self, q: Tensor, k: Tensor, mask: Optional[BoolTensor] = None
    ) -> Tensor:
        score = q.matmul(k.transpose(-1, -2))
        if self.apply_rel_dist is not None:
            # score_{ij} = <q_i, k_j> + s_{ij}
            klen = k.size(2)
            # idx.shape = [klen, klen]
            # idx[i][j] = i-j
            idx = pt.arange(klen, dtype=pt.long, device=k.device)
            idx = idx.view(-1, 1) - idx.view(1, -1)
            # idx[i][j] = |i-j|
            idx = idx.abs_()
            # torch.gather requires consistent shape
            # idx.shape = [batch, n_head, klen, klen]
            idx = idx.view(1, 1, klen, klen).expand(
                k.size(0), k.size(1), -1, -1
            )
            # dist representation r for attention
            # r.shape = [1, klen, d_hidden]
            r = pt.arange(klen, dtype=pt.float, device=k.device).view(1, -1)
            r = linear(self.posemb(r), self._pos_proj_weight)
            # r.shape = [1, n_head, klen, d_head]
            r = self._split_head(r)
            if self.apply_rel_dist == "dot":
                # s_{ij} = <r_{|i-j|}, (q_i+k_j)>
                #        = <q_i, r_{|i-j|}> + <r_{|i-j|}, k_j>

                # qr_{ij} = <q_i, r_j>
                # qr'_{ij} = qr_{i,idx[i][j]} = qr_{i,|i-j|}
                qr = q.matmul(r.transpose(-1, -2))
                qr = qr.gather(dim=-1, index=idx)
                # rk_{ij} = <r_i, k_j>
                # rk'_{ij} = rk_{idx[i][j], j} = rk_{|i-j|, j}
                rk = r.matmul(k.transpose(-1, -2))
                rk = rk.gather(dim=-2, index=idx)
                # s_{ij} = qr_{i,|i-j|} + rk_{|i-j|,j}
                s = qr + rk
            else:
                # transformer-xl style: https://arxiv.org/abs/1901.02860
                # s_{ij} = <q_i, r_{|i-j|}> + <u, k_j> + <v, r_{|i-j|}>
                #      u = _content_bias_weight
                #      v = _position_bias_weight

                # qr_{ij} = <q_i, r_j>
                # qr'_{ij} = qr_{i,idx[i][j]} = qr_{i,|i-j|}
                qr = q.matmul(r.transpose(-1, -2))
                qr = qr.gather(dim=-1, index=idx)
                # rk_{ij} = <v, r_i> + <u, k_j>
                # rk'_{ij} = rk_{idx[i][j], j} = rk_{|i-j|, j}
                rk = r.matmul(
                    self._content_bias_weight.transpose(-1, -2)
                ) + self._position_bias_weight.matmul(k.transpose(-1, -2))
                rk = rk.gather(dim=-2, index=idx)
                # s_{ij} = qr_{i,|i-j|} + rk_{|i-j|, j}
                s = qr + rk
            # add relative positional bias to content-based attention score
            score = score + s

        score = self._apply_mask(score, mask)
        score = score.div(math.sqrt(self.d_head)).div(self.temperature)
        score = score.softmax(dim=-1)
        score = self.dropout(score)
        return score

    def _compute_attn_output(self, score: Tensor, v: Tensor) -> Tensor:
        v = self._merge_head(score.matmul(v))
        v = linear(v, self._out_proj_weight, self._out_proj_bias)
        return v

    def forward(
        self, x: Tensor, *, mask: Optional[BoolTensor] = None
    ) -> Tensor:
        q, k, v = self._compute_qkv(x)
        score = self._compute_attn_score(q, k, mask)
        v = self._compute_attn_output(score, v)
        return v


class GroupSelfAttention(SelfAttention):
    """
    Self-attention module with q,k from the same input tensor. The input tensor
    is the concatenation of `n_groups` of slightly different feature maps. Thus
    the projections are 1x1 group convolutions.

    *NOTE*: d_qk, d_hidden, n_head must be divisible by n_groups
    """

    def __init__(
        self,
        d_qk: int,
        d_value: int,
        n_head: int = 1,
        n_groups: int = 1,
        bias: bool = True,
        bidirectional: bool = False,
        apply_rel_dist: Optional[str] = None,
        share_values: bool = False,
        dropout: float = 0.0,
        temperature: float = 1.0,
    ):
        try:
            assert d_qk % n_groups == 0
            assert d_value % n_groups == 0
            assert n_head % n_groups == 0
        except AssertionError:
            raise ValueError(
                "d_qk, d_hidden, n_head must be divisible by n_groups"
            )
        if d_qk // n_groups > d_value:
            raise ValueError(
                "dimension of each group should not exceed d_value"
            )
        self.d_qk = d_qk
        self.d_value = d_value
        self.n_groups = n_groups

        super(GroupSelfAttention, self).__init__(
            d_hidden=d_value,
            n_head=n_head,
            bias=bias,
            bidirectional=bidirectional,
            apply_rel_dist=apply_rel_dist,
            share_values=share_values,
            dropout=dropout,
            temperature=temperature,
        )

    def _reset_parameters(self):
        del self._qk_proj_weight
        self._qk_proj_weight = nn.Parameter(
            Tensor(2 * self.d_hidden, self.d_qk // self.n_groups, 1)
        )
        super(GroupSelfAttention, self)._reset_parameters()

    def _compute_qkv(
        self, value: Tensor, shape: Tensor
    ) -> Tuple[Tensor, Tensor, Tensor]:
        qk = (
            conv1d(
                input=shape.transpose(1, 2),
                weight=self._qk_proj_weight,
                bias=self._qk_proj_bias,
                groups=self.n_groups,
            )
            .transpose(1, 2)
            .contiguous()
        )
        qk = qk.chunk(self.n_groups * 2, dim=-1)
        q = self._split_head(pt.cat(qk[0::2], dim=-1))
        k = self._split_head(pt.cat(qk[1::2], dim=-1))
        v = linear(value, self._value_proj_weight, self._value_proj_bias)
        if self.share_values:
            v = pt.stack([v] * self.n_head, dim=1)
        else:
            v = self._split_head(v)
        return q, k, v

    def forward(
        self,
        value: Tensor,
        shape: Tensor,
        *,
        mask: Optional[BoolTensor] = None,
    ) -> Tensor:
        q, k, v = self._compute_qkv(value, shape)
        score = self._compute_attn_score(q, k, mask)
        v = self._compute_attn_output(score, v)
        return v


class CompositeAttention(GroupSelfAttention):
    """
    Self-attention module with groups that has a regressive component.
    """

    def _compute_qkv(
        self, value: Tensor, shape: Tensor, estimate: Tensor
    ) -> Tuple[Tensor, Tensor, Tensor, Tensor]:
        q, k, v = super(CompositeAttention, self)._compute_qkv(value, shape)
        v = v.roll(-1, dim=2)
        e = self._split_head(e)
        return q, k, v, e

    def _compute_attn_output(
        self, score: Tensor, v: Tensor, e: Tensor
    ) -> Tensor:
        history_score = score.tril(diagonal=-1)
        current_score = score.diagonal(dim1=-2, dim2=-1).unsqueeze(dim=-1)
        v = history_score @ v + current_score * e
        v = self._merge_head(v)
        v = linear(v, self._out_proj_weight, self._out_proj_bias)
        return v

    def forward(
        self,
        value: Tensor,
        shape: Tensor,
        estimate: Tensor,
        mask: Optional[BoolTensor] = None,
    ) -> Tensor:
        q, k, v, e = self._compute_qkv(value, shape, estimate)
        score = self._compute_attn_score(q, k, mask)
        v = self._compute_attn_output(score, v, e)
        return v



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import inspect
from typing import Optional

import torch
import torch.nn as nn


def get_module_forward_input_names(module: nn.Module):
    params = inspect.signature(module.forward).parameters
    param_names = [k for k, v in params.items() if not str(v).startswith("*")]
    return param_names


def weighted_average(
    x: torch.Tensor, weights: Optional[torch.Tensor] = None, dim=None
) -> torch.Tensor:
    """
    Computes the weighted average of a given tensor across a given dim, masking
    values associated with weight zero,

    meaning instead of `nan * 0 = nan` you will get `0 * 0 = 0`.

    Parameters
    ----------
    x
        Input tensor, of which the average must be computed.
    weights
        Weights tensor, of the same shape as `x`.
    dim
        The dim along which to average `x`

    Returns
    -------
    Tensor:
        The tensor with values averaged along the specified `dim`.
    """
    if weights is not None:
        weighted_tensor = torch.where(
            weights != 0, x * weights, torch.zeros_like(x)
        )
        sum_weights = torch.clamp(
            weights.sum(dim=dim) if dim else weights.sum(), min=1.0
        )
        return (
            weighted_tensor.sum(dim=dim) if dim else weighted_tensor.sum()
        ) / sum_weights
    else:
        return x.mean(dim=dim)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.


from typing import Iterator, List

import numpy as np
import pandas as pd

from src.gluonts.core.component import validated
from src.gluonts.model.forecast import SampleForecast
from src.gluonts.mx import Tensor


class TemporalHierarchy:
    @validated()
    def __init__(self, agg_multiples: List[int], freq_strs: List[str]):
        self.num_leaves = agg_multiples[0]
        self.agg_multiples = agg_multiples
        self.freq_strs = freq_strs
        self.agg_mat = TemporalHierarchy._get_agg_mat(
            agg_multiples=agg_multiples
        )

    @property
    def _adj_mat(self):
        """
        Adjacency matrix for the temporal hierarchy.

        This is created in the following way:
            * First get indices of source vertices and destination vertices:
            source vertices are the aggregated nodes and destination vertices are their children in that order;
            e.g., for 15-min hierarchy, `source_ix` = [0, 1, 2] and `dest_ix` = [[1, 2], [3, 4], [5, 6]],
            i.e., 0 --> [1, 2], 1 --> [3, 4], 2 --> [5, 6].
            * Once we have these two lists, we can simply do `adj[s, t] = 1`, for `s` in `source_ix` and `t` `dest_ix`.
            * This is made symmetric by adding `adj[t, s] = 1`.

        :return:
        """
        num_nodes_running_sum = np.cumsum(self.num_nodes_per_level)

        # for 15-min hierarchy: [[1, 2], [3, 4, 5, 6]] (we ignore the root).
        node_ix_level_wise = [
            list(range(i, j))
            for i, j in zip(
                num_nodes_running_sum[:-1], num_nodes_running_sum[1:]
            )
        ]

        dest_ix = []
        for n, nodes_ix in zip(
            self.num_nodes_per_level[:-1], node_ix_level_wise
        ):
            # split `node_ix` list into `n` equal parts;
            # this relies on the definition of temporal hierarchy: equal splits during disaggregation!
            # for 15-min hierarchy: [1, 2] --> [1, 2]; [3, 4, 5, 6] --> [3, 4], [5, 6]
            dest_ix.extend(np.array_split(nodes_ix, n))

        total_num_nodes = np.sum(self.num_nodes_per_level)
        source_ix = list(range(total_num_nodes - self.num_leaves))

        adj = np.zeros((total_num_nodes, total_num_nodes))
        for s, t in zip(source_ix, dest_ix):
            adj[s, t] = 1
            adj[t, s] = 1

        return adj

    @property
    def _hierarchical_adj_mat(self):
        mat_accumulate = np.triu(self._adj_mat)
        mat_distribute = np.tril(self._adj_mat)
        col_sums = mat_distribute.sum(axis=0)
        mat_distribute[:, col_sums > 0] = mat_distribute[
            :, col_sums > 0
        ] / mat_distribute[:, col_sums > 0].sum(axis=0, keepdims=True)
        I = np.eye(mat_accumulate.shape[0])

        return (mat_accumulate + mat_distribute + I) / 3.0

    def adj_mat(self, option: str = "hierarchical"):
        assert option in [
            "standard_plus_identity",  # A + I
            "normalized_with_identity",  # D^{-1} (A + I), where D is the diagonal matrix of (A + I)
            "symmetric_normalized_with_identity",  # D^{-1/2} (A + I) D^{-1/2}, D is the diagonal matrix of (A + I)
            "standard",  # A
            "normalized",  # D^{-1} A, where D is the diagonal matrix of A
            "symmetric_normalized",  # D^{-1/2} A D^{-1/2}, D is the diagonal matrix of A
            "hierarchical",
        ]
        if option == "standard":
            return self._adj_mat
        if option == "standard_plus_identity":
            A = self._adj_mat
            return A + np.eye(A.shape[0])
        if option == "normalized":
            A = self._adj_mat
            D = np.diag(A.sum(axis=1))
            D_inv = np.linalg.inv(D)
            A_normalized = np.matmul(D_inv, A)
            return A_normalized
        if option == "normalized_with_identity":
            A = self._adj_mat
            A_hat = A + np.eye(A.shape[0])
            D = np.diag(A_hat.sum(axis=1))
            D_inv = np.linalg.inv(D)
            A_normalized_with_identity = np.matmul(D_inv, A_hat)
            return A_normalized_with_identity
        if option == "symmetric_normalized":
            A = self._adj_mat
            sqrt_D = np.diag(np.sqrt(A.sum(axis=1)))
            D_inv_sqrt = np.linalg.inv(sqrt_D)
            A_symmetric_normalized = np.matmul(
                D_inv_sqrt, np.matmul(A, D_inv_sqrt)
            )
            return A_symmetric_normalized
        if option == "symmetric_normalized_with_identity":
            A = self._adj_mat
            A_hat = A + np.eye(A.shape[0])
            sqrt_D = np.diag(np.sqrt(A_hat.sum(axis=1)))
            D_inv_sqrt = np.linalg.inv(sqrt_D)
            A_symmetric_normalized_with_identity = np.matmul(
                D_inv_sqrt, np.matmul(A_hat, D_inv_sqrt)
            )
            return A_symmetric_normalized_with_identity
        if option == "hierarchical":
            return self._hierarchical_adj_mat

    @property
    def num_nodes_per_level(self):
        return [
            self.num_leaves // agg_multiple
            for agg_multiple in self.agg_multiples
        ]

    @property
    def nodes(self):
        """
        Returns node structure representing the hierarchy as defined in the hts package.
        Examples:
            ["1D", "8H", "1H"] --> [3, [8, 8, 8]]
            ["1H", "30min", "1min"] --> [2, [30, 30]]

        Note that the nodes structure does not work if any node is *not* formed by grouping its immediate descendants,
        e.g., ["6H", "4H", "2H"].

        To know the exact structure of nodes see the help:
        Hierarchical: https://stackoverflow.com/questions/13579292/how-to-use-hts-with-multi-level-hierarchies
        Grouped: https://robjhyndman.com/hyndsight/gts/

        :return:
        """
        agg_multiples = np.array(self.agg_multiples)
        num_children = agg_multiples[:-1] // agg_multiples[1:]
        num_nodes_at_agg_levels = self.num_nodes_per_level[:-1]

        nodes = [
            [nc] * num_nodes if num_nodes > 1 else nc
            for nc, num_nodes in zip(num_children, num_nodes_at_agg_levels)
        ]
        return nodes

    @staticmethod
    def _get_agg_mat(agg_multiples: List[int]):
        num_leaves = agg_multiples[0]

        rows = []
        for agg_multiple in agg_multiples:
            num_rows = num_leaves // agg_multiple
            for i in range(num_rows):
                row = np.zeros(num_leaves)
                row[i * agg_multiple : (i + 1) * agg_multiple] = 1
                rows.append(row)

        return np.vstack(rows)


def _check_freqs(deltas: List[pd.Timedelta]):
    """
    Checks if all lower frequencies can be aggregated by highest frequency.
    """
    assert all(
        [d % deltas[-1] == pd.Timedelta("0H") for d in deltas]
    ), "Lower frequencies are not multiples of highest frequency."


def freqs_to_agg_mulitples(freq_strs: List[str]) -> List[int]:
    """
    Returns aggregation multiples that are used to construct aggregation
    matrix.

    Parameters
    ----------
    freq_strs : List[str]
        ordered list of frequencies. e.g ["1D", "8H", "1H"]
    """
    deltas = [pd.Timedelta(freq) for freq in freq_strs]
    _check_freqs(deltas)
    return [d // deltas[-1] for d in deltas]


def to_TemporalHierarchy(freq_strs: List[str]) -> TemporalHierarchy:
    agg_multiples = freqs_to_agg_mulitples(freq_strs)
    return TemporalHierarchy(agg_multiples=agg_multiples, freq_strs=freq_strs)


def agg_series(seq: Tensor, agg_multiple: int):
    batch_size, seq_length = seq.shape[0:2]

    return seq.reshape(
        batch_size,
        seq_length // agg_multiple,
        agg_multiple,
        -1,
    ).sum(axis=2)


def unpack_forecasts(
    forecast_at_all_levels_it: Iterator[SampleForecast],
    temporal_hierarchy: TemporalHierarchy,
    target_temporal_hierarchy: TemporalHierarchy,
    num_samples: int = 100,
) -> Iterator[SampleForecast]:
    """

    :param forecast_at_all_levels_it:
        Each element of the iterator has `SampleForecast` object with samples of shape:
            (`num_samples`, `prediction_length_at_root`, `total_num_nodes`),
        where the order of the last axis is level-wise left-right ordering of the hierarchy.
    :param temporal_hierarchy:
    :param num_samples:
        Used only for asserting shapes.
    :return:
    This will get the forecasts in this order:
        (ts_1_6M, ts_1_2M, ts_1_1M, ts_2_6M, ts_2_2M, ts_2_1M, ...)

    """

    num_nodes_per_level = temporal_hierarchy.num_nodes_per_level

    for forecast in forecast_at_all_levels_it:
        # We iterate over each time series and unpack forecasts for that time series at different
        # time granularites at the same time.
        samples_at_all_levels = forecast.samples
        assert num_samples == forecast.num_samples

        start_ix = 0
        for num_nodes, agg_freq, agg_multiple in zip(
            num_nodes_per_level,
            temporal_hierarchy.freq_strs,
            temporal_hierarchy.agg_multiples,
        ):
            end_ix = start_ix + num_nodes

            forecast_start_date = pd.Period(
                str(forecast.start_date),
                freq=(agg_multiple * forecast.start_date.freq).freqstr,
            )
            agg_forecast = SampleForecast(
                samples=samples_at_all_levels[..., start_ix:end_ix].reshape(
                    num_samples, -1
                ),
                start_date=forecast_start_date,
                item_id=forecast.item_id,
                info=forecast.info,
            )
            start_ix = end_ix

            if agg_multiple in target_temporal_hierarchy.agg_multiples:
                yield agg_forecast

        # Make sure we used all the samples!
        assert start_ix == samples_at_all_levels.shape[-1]


def get_ts_at_all_levels(
    ts_it: Iterator[pd.Series],
    temporal_hierarchy: TemporalHierarchy,
    prediction_length: int,
    target_temporal_hierarchy: TemporalHierarchy,
):
    for ts in ts_it:
        for agg_freq, agg_multiple in zip(
            temporal_hierarchy.freq_strs, temporal_hierarchy.agg_multiples
        ):
            if agg_multiple in target_temporal_hierarchy.agg_multiples:
                # We need to align the forecast start date for all aggregation levels.
                # This is possible if the time series length is multiple of `prediction_length`.
                ts = ts[-(len(ts) // prediction_length) * prediction_length :]

                agg_ts_values = ts.values.reshape(
                    len(ts) // agg_multiple, agg_multiple
                ).sum(axis=1)

                # Start date is same for the aggregated frequency:
                # For example if you aggregate 5-minutes data, that starts at 00:10, to hourly then the index of hourly
                # data will be 00:10, 01:10, 02:10, ...
                agg_ts_start_date = ts.index[0]
                agg_ts_index = pd.period_range(
                    start=agg_ts_start_date,
                    periods=len(agg_ts_values),
                    freq=(agg_multiple * ts.index[0].freq).freqstr,
                )
                agg_ts = pd.Series(agg_ts_values, index=agg_ts_index)

                yield agg_ts


def naive_reconcilation_mat(S: np.ndarray, nodes: List):
    """
    Returns the (average) reconciliation matrix that reconciles forecasts from all levels.
    In particular, it first computes mapping matrices `G` to reconcile base forecasts of any chosen level.
    The G matrix is the one described in the book: https://otexts.com/fpp2/mapping-matrices.html
    After obtining such `G` matrices for each level (G_1, G_2, ..., G_k), it computes the final `projection` or the
    `reconciliation` matrix as:
                P = S (G_1 + G_2 + ... + G_k) / k.
    This way it would use the un-reconciled forecasts of all nodes in the hierarchy to arrive at the final coherent
    forecasts.
    Parameters
    ----------
    S
        Summation or aggregation matrix. Shape: (total_num_time_series, num_base_time_series)
    nodes
        Node structure representing the hierarchy as defined in the hts package.
        To know the exact structure of nodes see the help:
        Hierarchical: https://stackoverflow.com/questions/13579292/how-to-use-hts-with-multi-level-hierarchies
        Grouped: https://robjhyndman.com/hyndsight/gts/
    Returns
    -------
    Numpy array of shape: (total_num_time_series, total_num_time_series)
    """
    num_ts, num_bottom_series = S.shape
    num_nodes_per_level = [1] + [np.asscalar(np.sum(n)) for n in nodes]
    cum_num_nodes_per_level = np.cumsum(num_nodes_per_level)

    def mapping_matrix_at_level(level: int):
        start_ix = 0 if level == 0 else cum_num_nodes_per_level[level - 1]
        end_ix = cum_num_nodes_per_level[level]

        M = np.zeros((num_bottom_series, num_ts))
        M[:, start_ix:end_ix] = S[start_ix:end_ix, :].transpose()

        # equal proportions
        row_sum = M[:, start_ix:end_ix].sum(axis=0)
        M[:, start_ix:end_ix] = M[:, start_ix:end_ix] / row_sum[None, :]
        return M

    mapping_matrices = np.array(
        [
            mapping_matrix_at_level(level=level)
            for level in range(len(cum_num_nodes_per_level))
        ]
    )

    mean_mapping_matrix = np.mean(mapping_matrices, axis=0)
    reconciliation_mat = np.matmul(S, mean_mapping_matrix)

    return reconciliation_mat



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.


from itertools import islice
from pathlib import Path
import os


from src.gluonts.dataset.repository import get_dataset
from src.gluonts.nursery.autogluon_tabular import TabularEstimator
from src.gluonts.model.predictor import Predictor


def run_example():
    dataset = get_dataset("electricity")
    serialize_path = Path("GluonTSTabularPredictor")
    estimator = TabularEstimator(
        freq="H",
        prediction_length=24,
        time_limit=600,  # ten minutes for training
        disable_auto_regression=True,  # makes prediction faster, but potentially less accurate
        last_k_for_val=24,  # split the last 24 targets from each time series to be the validation data
        quantiles_to_predict=[0.1, 0.5, 0.9],
    )

    n_train = 5

    training_data = list(islice(dataset.train, n_train))

    predictor = estimator.train(training_data=training_data)

    os.makedirs(serialize_path, exist_ok=True)
    predictor.serialize(serialize_path)
    predictor = None
    # the quantiles_to_predict parameters should be List[str] type
    predictor = Predictor.deserialize(serialize_path)
    forecasts = list(predictor.predict(training_data))
    print(forecasts)


if __name__ == "__main__":
    run_example()



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import cast, List, Literal, Optional
import numpy as np
import numpy.typing as npt
import pytorch_lightning as pl
import torch
from lightkit.data import DataLoader
from torch import nn
from torch.utils.data import TensorDataset
from tsbench.config import Config, EnsembleConfig
from tsbench.evaluations.tracking import EnsembleTracker
from tsbench.surrogate.torch.deepset import DeepSetModel
from ._base import OutputNormalization, Surrogate
from ._factory import register_ensemble_surrogate
from .torch import DeepSetLightningModule, ListMLELoss
from .transformers import EnsembleConfigTransformer


@register_ensemble_surrogate("deepset")
class DeepSetSurrogate(Surrogate[EnsembleConfig]):
    """
    The DeepSet surrogate is similar to the MLP surrogate but makes predictions
    for ensembles rather than single models.

    Currently, it does not support the use of dataset features.
    """

    trainer_: pl.Trainer
    models_: List[nn.Module]

    def __init__(
        self,
        tracker: EnsembleTracker,
        objective: Literal["regression", "ranking"] = "ranking",
        discount: Optional[
            Literal["logarithmic", "linear", "quadratic"]
        ] = "linear",
        hidden_layer_sizes: Optional[List[int]] = None,
        weight_decay: float = 0.01,
        dropout: float = 0.0,
        predict: Optional[List[str]] = None,
        output_normalization: OutputNormalization = None,
        impute_simulatable: bool = False,
    ):
        """
        Args:
            tracker: A tracker that can be used to impute latency and number of model parameters
                into model performances. Also, it is required for some input features.
            objective: The optimization objective for the XGBoost estimators.
            discount: The discount to apply for the ranking loss. If provided, it focuses on
                correctly predicting the top values.
            hidden_layer_sizes: The dimensions of the hidden layers. Defaults to two hidden layers
                of size 32.
            weight_decay: The weight decay to apply during optimization.
            dropout: The dropout probability of dropout layers applied after every activation
                function.
            predict: The metrics to predict. All if not provided.
            output_normalization: The type of normalization to apply to the features of each
                dataset independently. `None` applies no normalization, "quantile" applies quantile
                normalization, and "standard" transforms data to have zero mean and unit variance.
            impute_simulatable: Whether the tracker should impute latency and number of model
                parameters into the returned performance object.
        """

        super().__init__(
            tracker, predict, output_normalization, impute_simulatable
        )

        self.use_ranking = objective == "ranking"
        self.hidden_layer_sizes = hidden_layer_sizes or [32, 32]
        self.weight_decay = weight_decay
        self.dropout = dropout

        self.config_transformer = EnsembleConfigTransformer()

        if objective == "regression":
            self.loss = nn.MSELoss()
        elif objective == "ranking":
            self.loss = ListMLELoss(discount=discount)

    @property
    def required_cpus(self) -> int:
        return 4

    def _fit(
        self, X: List[Config[EnsembleConfig]], y: npt.NDArray[np.float32]
    ) -> None:
        # Fit transformers to infer dimensionality
        X_numpy_list = self.config_transformer.fit_transform(X)
        X_numpy = np.concatenate(X_numpy_list)
        X_lengths_numpy = np.array([x.shape[0] for x in X_numpy_list])

        input_dim = len(self.config_transformer.feature_names_)
        output_dim = y.shape[1]

        # For initializing data, we prepare group IDs for the datasets
        mapping = {d: i for i, d in enumerate({x.dataset for x in X})}

        # For each output variable, we need to train a separate model
        self.models_ = []
        for i in range(output_dim):
            model = self._init_model(input_dim)
            module = DeepSetLightningModule(
                model, self.loss, self.weight_decay
            )

            # Train on output variable i
            dataset = TensorDataset(
                torch.from_numpy(X_numpy).float(),
                torch.from_numpy(X_lengths_numpy).long(),
                torch.from_numpy(y[:, i : i + 1]).float(),
                torch.as_tensor(
                    [mapping[x.dataset] for x in X], dtype=torch.long
                ),
            )
            train_loader = DataLoader(dataset, batch_size=len(dataset))
            self._trainer.fit(module, train_dataloaders=train_loader)

            # Add to models
            self.models_.append(model)

    def _predict(
        self, X: List[Config[EnsembleConfig]]
    ) -> npt.NDArray[np.float32]:
        # Get data
        X_numpy_list = self.config_transformer.transform(X)
        X_numpy = np.concatenate(X_numpy_list)
        X_lengths_numpy = np.array([x.shape[0] for x in X_numpy_list])

        dataset = TensorDataset(
            torch.from_numpy(X_numpy).float(),
            torch.from_numpy(X_lengths_numpy).long(),
        )
        test_loader = DataLoader(dataset, batch_size=len(dataset))

        # Run prediction
        predictions = []
        for model in self.models_:
            module = DeepSetLightningModule(model, self.loss)
            out = cast(
                List[torch.Tensor], self._trainer.predict(module, test_loader)
            )
            predictions.append(out[0].numpy())

        return np.concatenate(predictions, axis=-1)

    @property
    def _trainer(self) -> pl.Trainer:
        return pl.Trainer(
            max_epochs=1000,
            logger=False,
            enable_checkpointing=False,
            enable_model_summary=False,
            enable_progress_bar=False,
            gpus=int(torch.cuda.is_available()),
        )

    def _init_model(self, input_dim: int) -> nn.Module:
        return DeepSetModel(
            input_dim,
            self.hidden_layer_sizes[-1],
            1,
            self.hidden_layer_sizes,
            self.hidden_layer_sizes,
            self.dropout,
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import logging

import numpy as np
import pandas as pd

from src.gluonts.dataset.common import ListDataset, DataEntry, Dataset


class Grouper:
    # todo the contract of this grouper is missing from the documentation, what it does when, how it pads values etc
    def __init__(
        self,
        fill_value: float = 0.0,
        max_target_dim: int = None,
        align_data: bool = True,
        num_test_dates: int = None,
    ) -> None:
        self.fill_value = fill_value

        self.first_timestamp = pd.Timestamp(2200, 1, 1, 12)
        self.last_timestamp = pd.Timestamp(1800, 1, 1, 12)
        self.frequency = None
        self.align_data = align_data
        self.max_target_length = 0
        self.num_test_dates = num_test_dates
        self.max_target_dimension = max_target_dim

    def __call__(self, dataset: Dataset) -> Dataset:
        self._preprocess(dataset)
        return self._group_all(dataset)

    def _group_all(self, dataset: Dataset) -> Dataset:
        if self.align_data:
            funcs = {"target": self._align_data_entry}

        if self.num_test_dates is None:
            grouped_dataset = self._prepare_train_data(dataset, funcs)
        else:
            grouped_dataset = self._prepare_test_data(dataset)
        return grouped_dataset

    def to_ts(self, data: DataEntry):
        return pd.Series(
            data["target"],
            index=pd.date_range(
                start=data["start"],
                periods=len(data["target"]),
                freq=data["start"].freq,
            ),
        )

    def _align_data_entry(self, data: DataEntry) -> DataEntry:
        d = data.copy()
        # fill target invidually if we want to fill all of them, we should use a dataframe
        ts = self.to_ts(data)
        d["target"] = ts.reindex(
            pd.date_range(
                start=self.first_timestamp,
                end=self.last_timestamp,
                freq=d["start"].freq,
            ),
            fill_value=ts.mean(),
        )
        d["start"] = self.first_timestamp
        return d

    def _preprocess(self, dataset: Dataset) -> None:
        """
        The preprocess function iterates over the dataset to gather data that
        is necessary for grouping.

        This includes:
             1. Storing first/last timestamp in the dataset
             2. Aligning time series
             3. Calculating groups
        """
        for data in dataset:
            timestamp = data["start"]
            self.first_timestamp = min(self.first_timestamp, timestamp)

            self.frequency = (
                timestamp.freq if self.frequency is None else self.frequency
            )
            self.last_timestamp = max(
                self.last_timestamp,
                timestamp + len(data["target"]) * self.frequency,
            )

            # todo
            self.max_target_length = max(
                self.max_target_length, len(data["target"])
            )
        logging.info(
            f"first/last timestamp found: {self.first_timestamp}/{self.last_timestamp}"
        )

    def _prepare_train_data(self, dataset, funcs):
        logging.info("group training time-series to datasets")
        grouped_data = {}
        for key in funcs.keys():
            grouped_entry = [funcs[key](data)[key] for data in dataset]

            # we check that each time-series has the same length
            assert (
                len(set([len(x) for x in grouped_entry])) == 1
            ), f"alignement did not work as expected more than on length found: {set([len(x) for x in grouped_entry])}"
            grouped_data[key] = np.array(grouped_entry)
        if self.max_target_dimension is not None:
            # targets are often sorted by incr amplitude, use the last one when restricted number is asked
            grouped_data["target"] = grouped_data["target"][
                -self.max_target_dimension :, :
            ]
        grouped_data["item_id"] = "all_items"
        grouped_data["start"] = self.first_timestamp
        grouped_data["feat_static_cat"] = [0]
        return ListDataset(
            [grouped_data], freq=self.frequency, one_dim_target=False
        )

    def _prepare_test_data(self, dataset):
        logging.info("group test time-series to datasets")

        def left_pad_data(data: DataEntry):
            ts = self.to_ts(data)
            filled_ts = ts.reindex(
                pd.date_range(
                    start=self.first_timestamp,
                    end=ts.index[-1],
                    freq=data["start"].freq,
                ),
                fill_value=0.0,
            )
            return filled_ts.values

        grouped_entry = [left_pad_data(data) for data in dataset]
        grouped_entry = np.array(grouped_entry)

        split_dataset = np.split(grouped_entry, self.num_test_dates)

        all_entries = list()
        for dataset_at_test_date in split_dataset:
            grouped_data = dict()
            assert (
                len(set([len(x) for x in dataset_at_test_date])) == 1
            ), "all test time-series should have the same length"
            grouped_data["target"] = np.array(
                list(dataset_at_test_date), dtype=np.float32
            )
            if self.max_target_dimension is not None:
                grouped_data["target"] = grouped_data["target"][
                    -self.max_target_dimension :, :
                ]
            grouped_data["item_id"] = "all_items"
            grouped_data["start"] = self.first_timestamp
            grouped_data["feat_static_cat"] = [0]
            all_entries.append(grouped_data)

        return ListDataset(
            all_entries, freq=self.frequency, one_dim_target=False
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from collections import OrderedDict
import torch.nn as nn
import torch


def load_weights(model: nn.Module, path_to_weights: str) -> nn.Module:
    ckpt = torch.load(
        path_to_weights, map_location=lambda storage, loc: storage
    )

    # this seems to be necessary
    for key in ckpt["state_dict"].copy():
        if ("loss" in key) or ("crps" in key) or ("quantile_width" in key):
            ckpt["state_dict"].pop(key)

    new_state_dict = OrderedDict()
    for k, v in ckpt["state_dict"].items():
        name = k.replace("model.", "")
        new_state_dict[name] = v

    model.load_state_dict(new_state_dict)
    return model



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import List, Optional

import torch
import torch.nn as nn

from src.gluonts.core.component import validated
from src.gluonts.torch.util import copy_parameters
from src.gluonts.torch.model.predictor import PyTorchPredictor
from src.gluonts.torch.modules.distribution_output import DistributionOutput
from src.gluonts.model.predictor import Predictor
from src.gluonts.dataset.field_names import FieldName
from src.gluonts.transform import (
    Transformation,
    Chain,
    InstanceSplitter,
    ExpectedNumInstanceSampler,
    ValidationSplitSampler,
    TestSplitSampler,
)

from pts.model.utils import get_module_forward_input_names
from pts import Trainer
from pts.model import PyTorchEstimator
from pts.modules import StudentTOutput

from .simple_feedforward_network import (
    SimpleFeedForwardTrainingNetwork,
    SimpleFeedForwardPredictionNetwork,
)


class SimpleFeedForwardEstimator(PyTorchEstimator):
    """
    SimpleFeedForwardEstimator shows how to build a simple MLP model predicting
    the next target time-steps given the previous ones.

    Given that we want to define a pytorch model trainable by SGD, we inherit the
    parent class `PyTorchEstimator` that handles most of the logic for fitting a
    neural-network.

    We thus only have to define:

    1. How the data is transformed before being fed to our model::

        def create_transformation(self) -> Transformation

    2. How the training happens::

        def create_training_network(self) -> nn.Module

    3. how the predictions can be made for a batch given a trained network::

        def create_predictor(
             self,
             transformation: Transformation,
             trained_net: nn.Module,
        ) -> Predictor


    Parameters
    ----------
    freq
        Time time granularity of the data
    prediction_length
        Length of the prediction horizon
    trainer
        Trainer object to be used (default: Trainer())
    num_hidden_dimensions
        Number of hidden nodes in each layer (default: [40, 40])
    context_length
        Number of time units that condition the predictions
        (default: None, in which case context_length = prediction_length)
    distr_output
        Distribution to fit (default: StudentTOutput())
    batch_normalization
        Whether to use batch normalization (default: False)
    mean_scaling
        Scale the network input by the data mean and the network output by
        its inverse (default: True)
    num_parallel_samples
        Number of evaluation samples per time series to increase parallelism during inference.
        This is a model optimization that does not affect the accuracy (default: 100)
    """

    @validated()
    def __init__(
        self,
        freq: str,
        prediction_length: int,
        trainer: Trainer = Trainer(),
        num_hidden_dimensions: Optional[List[int]] = None,
        context_length: Optional[int] = None,
        distr_output: DistributionOutput = StudentTOutput(),
        batch_normalization: bool = False,
        mean_scaling: bool = True,
        num_parallel_samples: int = 100,
    ) -> None:
        """
        Defines an estimator.

        All parameters should be serializable.
        """
        super().__init__(trainer=trainer)

        self.num_hidden_dimensions = (
            num_hidden_dimensions
            if num_hidden_dimensions is not None
            else list([40, 40])
        )
        self.prediction_length = prediction_length
        self.context_length = (
            context_length if context_length is not None else prediction_length
        )
        self.freq = freq
        self.distr_output = distr_output
        self.batch_normalization = batch_normalization
        self.mean_scaling = mean_scaling
        self.num_parallel_samples = num_parallel_samples

        self.train_sampler = ExpectedNumInstanceSampler(
            num_instances=1, min_future=prediction_length
        )
        self.validation_sampler = ValidationSplitSampler(
            min_future=prediction_length
        )

    # here we do only a simple operation to convert the input data to a form
    # that can be digested by our model by only splitting the target in two, a
    # conditioning part and a to-predict part, for each training example.
    # For a more complex transformation example, see the `pts.model.deepar`
    # transformation that includes time features, age feature, observed values
    # indicator, etc.
    def create_transformation(self) -> Transformation:
        return Chain([])

    def create_instance_splitter(self, mode: str):
        assert mode in ["training", "validation", "test"]
        instance_sampler = {
            "training": self.train_sampler,
            "validation": self.validation_sampler,
            "test": TestSplitSampler(),
        }[mode]

        return InstanceSplitter(
            target_field=FieldName.TARGET,
            is_pad_field=FieldName.IS_PAD,
            start_field=FieldName.START,
            forecast_start_field=FieldName.FORECAST_START,
            instance_sampler=instance_sampler,
            past_length=self.context_length,
            future_length=self.prediction_length,
            time_series_fields=[],  # [FieldName.FEAT_DYNAMIC_REAL]
        )

    # defines the network, we get to see one batch to initialize it.
    # the network should return at least one tensor that is used as a loss to minimize in the training loop.
    # several tensors can be returned for instance for analysis, see DeepARTrainingNetwork for an example.
    def create_training_network(
        self, device: torch.device
    ) -> SimpleFeedForwardTrainingNetwork:
        return SimpleFeedForwardTrainingNetwork(
            num_hidden_dimensions=self.num_hidden_dimensions,
            prediction_length=self.prediction_length,
            context_length=self.context_length,
            distr_output=self.distr_output,
            batch_normalization=self.batch_normalization,
            mean_scaling=self.mean_scaling,
        ).to(device)

    # we now define how the prediction happens given that we are provided a
    # training network.
    def create_predictor(
        self,
        transformation: Transformation,
        trained_network: nn.Module,
        device: torch.device,
    ) -> Predictor:
        prediction_splitter = self.create_instance_splitter("test")

        prediction_network = SimpleFeedForwardPredictionNetwork(
            num_hidden_dimensions=self.num_hidden_dimensions,
            prediction_length=self.prediction_length,
            context_length=self.context_length,
            distr_output=self.distr_output,
            batch_normalization=self.batch_normalization,
            mean_scaling=self.mean_scaling,
            num_parallel_samples=self.num_parallel_samples,
        ).to(device)

        copy_parameters(trained_network, prediction_network)
        input_names = get_module_forward_input_names(prediction_network)

        return PyTorchPredictor(
            input_transform=transformation + prediction_splitter,
            input_names=input_names,
            prediction_net=prediction_network,
            batch_size=self.trainer.batch_size,
            freq=self.freq,
            prediction_length=self.prediction_length,
            device=device,
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.


from typing import Optional, List, Tuple
from functools import reduce
from operator import mul, add
import math

import torch as pt
from torch import Tensor, LongTensor
from torch import nn
from torch.nn import init


class SinusoidalPositionalEmbedding(nn.Module):
    def __init__(self, d_embed: int):
        super(SinusoidalPositionalEmbedding, self).__init__()
        if d_embed % 2 > 0:
            raise ValueError(
                "sinusoidal embedding must have an even dimension"
            )
        self.d_embed = d_embed
        inv_freq = pt.arange(0, self.d_embed, 2, dtype=pt.float)
        inv_freq /= self.d_embed
        inv_freq *= -math.log(1e4)
        self.register_buffer("inv_freq", inv_freq.exp_())

    def forward(self, pos_seq: LongTensor) -> Tensor:
        pos_seq = pos_seq.unsqueeze(dim=-1).float() * self.inv_freq
        return pt.cat([pos_seq.sin(), pos_seq.cos()], dim=-1)


class LearnablePositionalEmbedding(nn.Module):
    def __init__(
        self,
        d_embed: int,
        max_len: int,
        hierarchy: Optional[List[Tuple[int, int]]] = None,
    ):
        super(LearnablePositionalEmbedding, self).__init__()
        if hierarchy is None:
            hierarchy = [(max_len, d_embed)]
        sub_shape, d_sub_embeds = zip(*hierarchy)
        assert (
            reduce(mul, sub_shape, 1) == max_len
        ), f"decomposed shape {sub_shape} inconsistent with max length {max_len}"
        assert (
            reduce(add, d_sub_embeds, 0) == d_embed
        ), f"decomposed embed dim {d_sub_embeds} inconsistent with overall embed dim {d_embed}"
        self.d_embed = d_embed
        self.d_sub_embeds = d_sub_embeds
        self.max_len = max_len
        self.sub_shape = sub_shape

        self._weights = nn.ParameterList(
            [
                nn.Parameter(Tensor(size, dim))
                for size, dim in zip(self.sub_shape, self.d_sub_embeds)
            ]
        )
        self._reset_parameters()

    def _reset_parameters(self):
        for weight in self._weights:
            init.xavier_normal_(weight)

    @property
    def n_levels(self):
        return len(self._weights)

    def forward(self, pos_seq: LongTensor):
        x = pos_seq.clone()
        embeddings = []
        for l in range(self.n_levels):
            size = self.sub_shape[l]
            weight = self._weights[l]
            index = x.remainder(size)
            embeddings.append(nn.functional.embedding(index, weight))
            x = x.div(size)
        if pt.any(x > 0):
            exceed_idx = [t[0].item() for t in pt.where(x > 0)]
            raise ValueError(
                f"Given positional index at {exceed_idx} exceeds the max sequence length {self.max_len}"
            )
        embeddings = pt.cat(embeddings, dim=-1)
        return embeddings



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from pathlib import Path
import yaml
from black import click
from cli.utils import (
    explode_key_values,
    iterate_configurations,
    run_sacred_script,
)
from tsbench.constants import DEFAULT_DATA_PATH, DEFAULT_EVALUATIONS_PATH
from ._main import analysis


@analysis.command(short_help="Evaluate the performance of recommended models.")
@click.option(
    "--experiment",
    required=True,
    help=(
        "The name of the experiment under which the individual training runs"
        " are grouped."
    ),
)
@click.option(
    "--config_path",
    required=True,
    help="The local path to the configuration file defining script options.",
)
@click.option(
    "--data_path",
    type=click.Path(exists=True),
    default=DEFAULT_DATA_PATH,
    show_default=True,
    help="The path where datasets are stored.",
)
@click.option(
    "--evaluations_path",
    type=click.Path(exists=True),
    default=DEFAULT_EVALUATIONS_PATH,
    show_default=True,
    help="The path where offline evaluations are stored.",
)
@click.option(
    "--nskip",
    default=0,
    show_default=True,
    help=(
        "The number of configurations to skip. Useful if some set of"
        " experiments failed."
    ),
)
def recommender(
    experiment: str,
    config_path: str,
    data_path: str,
    evaluations_path: str,
    nskip: int,
):
    """
    Evaluates the performance of the models proposed by a recommender on a
    hold-out dataset. Recommendations are made for each dataset by providing
    recommenders with the offline evaluations on the remaining datasets in the
    registry.

    This call runs the Sacred script for each provided configuration
    sequentially and returns only once all runs have completed.
    """
    with Path(config_path).open("r", encoding="utf-8") as f:
        content = yaml.safe_load(f)
        configs = explode_key_values("recommender", content)

    for configuration in iterate_configurations(configs, nskip):
        run_sacred_script(
            "recommender.py",
            experiment=experiment,
            data_path=data_path,
            evaluations_path=evaluations_path,
            **configuration,
        )
